import numpy as np
import pandas as pd
import piel

example_random_values = pd.Series(
    np.array(
        [
            "1111",
            "10010",
            "1011",
            "1001",
            "101",
            "11",
            "10000",
            "1101",
            "10001",
            "1100",
            "11010",
        ],
        dtype=np.int64,
    )
)

four_bits_array = piel.models.logic.electro_optic.bits_array_from_bits_amount(4)

basic_ideal_phase_map = piel.models.logic.electro_optic.linear_bit_phase_map(
    bits_amount=5, final_phase_rad=np.pi, initial_phase_rad=0
)


print(example_random_values)

phase_array = piel.models.logic.electro_optic.return_phase_array_from_data_series(
    data_series=example_random_values, phase_map=basic_ideal_phase_map
)
print(phase_array)
# phase_array = []
# for code_i in example_random_values.values:
#     phase = basic_ideal_phase_map[basic_ideal_phase_map.bits == str(code_i)].phase.values[0]
#     phase_array.append(phase)
# print(phase_array)

# print(basic_ideal_phase_map[basic_ideal_phase_map.bits == example_random_values].phase.values[0])
# print(basic_ideal_phase_map[basic_ideal_phase_map.bits == "1110"].phase.values[0])
# print(basic_ideal_phase_map)
# print(type(basic_ideal_phase_map.bits.iloc[-1]))
# print(basic_ideal_phase_map.bits.astype(dtype="<U4"))
# print(basic_ideal_phase_map)
# print(type(basic_ideal_phase_map.bits.iloc[-1]))
# print(four_bits_array)
# print(four_bits_array.dtype)



import pathlib
from piel.config import piel_path_types
from piel.file_system import return_path

__all__ = [
    "find_design_run",
]


def find_design_run(
    design_directory: piel_path_types,
    run_name: str | None = None,
) -> pathlib.Path:
    """
    For a given `design_directory`, the `openlane` output can be found in the `runs` subdirectory.

    They get sorted based on a reverse `list.sort()` method.

    # TODO docs
    """
    design_directory = return_path(design_directory)
    runs_design_directory = design_directory / "runs"
    # Convert to path so that it can be found and compared within design_directory
    all_runs_list = list(runs_design_directory.iterdir())
    if run_name is not None:
        run_path = runs_design_directory / run_name
        # Return this exact run
        if run_path in all_runs_list:
            # Check that the run exists
            pass
        else:
            raise ValueError(
                "Run: " + str(run_path) + "not found in " + str(all_runs_list)
            )
    else:
        # Take the latest design run if it exists
        if len(all_runs_list) > 0:
            all_runs_list.sort(reverse=True)
            latest_run = all_runs_list[0]
            run_path = latest_run
        else:
            # If there are no runs
            raise ValueError(
                "No OpenLane design runs were found in: " + str(runs_design_directory)
            )
    return run_path



__all__ = ["rename_gdsfactory_connections_to_spice", "convert_tuples_to_strings"]


def rename_gdsfactory_connections_to_spice(connections: dict):
    """
    We convert the connection connectivity of the gdsfactory netlist into names that can be integrated into a SPICE
    netlist. It iterates on each key value pair, and replaces each comma with an underscore.

    # TODO docs
    """
    spice_connections = {}
    for key, value in connections.items():
        new_key = key.replace(",", "_")
        new_value = value.replace(",", "_")
        spice_connections[new_key] = new_value
    return spice_connections


def convert_tuples_to_strings(tuple_list):
    result = {}
    for tpl in tuple_list:
        combined_string = "___".join(tpl).replace(",", "__")
        key, value = tpl
        result[key.split(",")[1]] = combined_string
    return result



import glob
import json
import openlane
import os
import pathlib
import shutil
import stat
import subprocess
import types
from typing import Literal
from .config import piel_path_types


__all__ = [
    "check_path_exists",
    "check_example_design",
    "copy_source_folder",
    "create_new_directory",
    "delete_path",
    "delete_path_list_in_directory",
    "get_files_recursively_in_directory",
    "permit_directory_all",
    "permit_script_execution",
    "setup_example_design",
    "read_json",
    "return_path",
    "run_script",
    "write_script",
]


def check_path_exists(
    path: piel_path_types,
    raise_errors: bool = False,
) -> bool:
    """
    Checks if a directory exists.

    Args:
        path(piel_path_types): Input path.

    Returns:
        directory_exists(bool): True if directory exists.
    """
    directory_exists = False
    path = return_path(path)
    if path.exists():
        directory_exists = True
    else:
        if raise_errors:
            raise ValueError("Path: " + str(path) + " does not exist.")
    return directory_exists


def check_example_design(
    design_name: str = "simple_design",
    designs_directory: piel_path_types | None = None,
) -> bool:
    """
    We copy the example simple_design from docs to the `/foss/designs` in the `iic-osic-tools` environment.

    Args:
        design_name(str): Name of the design to check.
        designs_directory(piel_path_types): Directory that contains the DESIGNS environment flag.
        # TODO

    Returns:
        None
    """
    if designs_directory is None:
        designs_directory = pathlib.Path(os.environ["DESIGNS"])

    design_folder = (
        designs_directory / design_name
    )  # TODO verify this copying operation
    return design_folder.exists()


def copy_source_folder(
    source_directory: piel_path_types,
    target_directory: piel_path_types,
) -> None:
    """
    Copies the files from a source_directory to a target_directory

    Args:
        source_directory(piel_path_types): Source directory.
        target_directory(piel_path_types): Target directory.

    Returns:
        None
    """
    source_directory = return_path(source_directory)
    target_directory = return_path(target_directory)
    if target_directory.exists():
        answer = input("Confirm deletion of: " + str(target_directory.resolve()))
        if answer.upper() in ["Y", "YES"]:
            shutil.rmtree(target_directory)
        elif answer.upper() in ["N", "NO"]:
            print(
                "Copying files now from: "
                + str(source_directory.resolve())
                + " to "
                + str(target_directory.resolve())
            )

    shutil.copytree(
        source_directory,
        target_directory,
        symlinks=False,
        ignore=None,
        copy_function=shutil.copy2,
        ignore_dangling_symlinks=False,
        dirs_exist_ok=False,
    )


def convert_list_to_path_list(
    input_list: list[piel_path_types],
) -> list[pathlib.Path]:
    """
    Converts a list of strings or pathlib.Path to a list of pathlib.Path.

    Args:
        input_list(list[piel_path_types]): Input list.

    Returns:
        output_list(list[pathlib.Path]): Output list.
    """
    output_list = []
    for item in input_list:
        item = return_path(item)
        output_list.append(item)
    return output_list


def create_new_directory(
    directory_path: str | pathlib.Path,
) -> None:
    """
    Creates a new directory.

    If the parents of the target_directory do not exist, they will be created too.

    Args:
        directory_path(str | pathlib.Path): Input path.

    Returns:
        None
    """
    directory_path = return_path(directory_path)

    # Check permissions of the parent to be able to create the directory
    parent_directory = directory_path.parent
    parent_directory_permissions = oct(parent_directory.stat().st_mode)

    # If permissions are not read, write and execute for all, we change them
    if parent_directory_permissions != "0o777":
        permit_directory_all(parent_directory)

    # Create the directory
    directory_path.mkdir(parents=True)


def delete_path(path: str | pathlib.Path) -> None:
    """
    Deletes a path.

    Args:
        path(str | pathlib.Path): Input path.

    Returns:
        None
    """
    path = return_path(path)
    if path.exists():
        if path.is_dir():
            shutil.rmtree(path)
        elif path.is_file():
            path.unlink()


def delete_path_list_in_directory(
    directory_path: piel_path_types,
    path_list: list,
    ignore_confirmation: bool = False,
    validate_individual: bool = False,
) -> None:
    """
    Deletes a list of files in a directory.

    Args:
        directory_path(piel_path_types): Input path.
        path_list(list): List of files.
        ignore_confirmation(bool): Ignore confirmation. Default: False.
        validate_individual(bool): Validate individual files. Default: False.

    Returns:
        None
    """
    directory_path = return_path(directory_path)
    path_list = convert_list_to_path_list(path_list)
    if validate_individual:
        if ignore_confirmation:
            for path in path_list:
                if path.exists():
                    delete_path(path)
        else:
            for path in path_list:
                if path.exists():
                    answer = input("Confirm deletion of: " + str(path))
                    if answer.upper() in ["Y", "YES"]:
                        delete_path(path)
                    elif answer.upper() in ["N", "NO"]:
                        print("Skipping deletion of: " + str(path))
    else:
        if ignore_confirmation:
            for path in path_list:
                if path.exists():
                    delete_path(path)
        else:
            answer = input("Confirm deletion of: " + str(path_list))
            if answer.upper() in ["Y", "YES"]:
                for path in path_list:
                    if path.exists():
                        delete_path(path)
            elif answer.upper() in ["N", "NO"]:
                print("Skipping deletion of: " + str(path_list))


def get_files_recursively_in_directory(
    path: piel_path_types,
    extension: str = "*",
):
    """
    Returns a list of files in a directory.

    Args:
        path(piel_path_types): Input path.
        extension(str): File extension.

    Returns:
        file_list(list): List of files.
    """
    path = return_path(path)
    file_list = []
    for x in os.walk(str(path.resolve())):
        for file_path in glob.glob(os.path.join(x[0], f"*.{extension}")):
            file_list.append(file_path)
    return file_list


def permit_script_execution(script_path: piel_path_types) -> None:
    """
    Permits the execution of a script.

    Args:
        script_path(piel_path_types): Script path.

    Returns:
        None
    """
    script = return_path(script_path)
    script.chmod(script.stat().st_mode | stat.S_IEXEC)


def permit_directory_all(directory_path: piel_path_types) -> None:
    """
    Permits a directory to be read, written and executed. Use with care as it can be a source for security issues.

    Args:
        directory_path(piel_path_types): Input path.

    Returns:
        None
    """
    directory_path = return_path(directory_path)
    try:
        directory_path.chmod(0o777)
    except PermissionError:
        print(
            UserWarning(
                "Could not change permissions of directory: "
                + str(directory_path.resolve())
                + " to 777. Your Python executable might not have the required permissions. Restructure your project directory so Python does not have to change permissions."
            )
        )


def read_json(path: piel_path_types) -> dict:
    """
    Reads a JSON file.

    Args:
        path(piel_path_types): Input path.

    Returns:
        json_data(dict): JSON data.
    """
    path = return_path(path)
    with open(path, "r") as json_file:
        json_data = json.load(json_file)
    return json_data


def return_path(input_path: piel_path_types) -> pathlib.Path:
    """
    Returns a pathlib.Path to be able to perform operations accordingly internally.

    This allows us to maintain compatibility between POSIX and Windows systems.

    Args:
        input_path(str): Input path.

    Returns:
        pathlib.Path: Pathlib path.
    """
    if type(input_path) == str:
        output_path = pathlib.Path(input_path)
    elif isinstance(input_path, pathlib.Path):
        output_path = input_path
    elif isinstance(input_path, types.ModuleType):
        # TODO Untested
        output_path = pathlib.Path(input_path.__file__) / ".."
    elif isinstance(input_path, os.PathLike):
        output_path = pathlib.Path(input_path)
    else:
        raise ValueError(
            "input_path: " + str(input_path) + " is of type: " + str(type(input_path))
        )
    return output_path


def run_script(script_path: piel_path_types) -> None:
    """
    Runs a script on the filesystem `script_path`.

    Args:
        script_path(piel_path_types): Script path.

    Returns:
        None
    """
    script = return_path(script_path)
    subprocess.run(str(script.resolve()), check=True, capture_output=True)


def setup_example_design(
    project_source: Literal["piel", "openlane"] = "piel",
    example_name: str = "simple_design",
) -> None:
    """
    We copy the example simple_design from docs to the `/foss/designs` in the `iic-osic-tools` environment.

    Args:
        project_source(str): Source of the project.
        example_name(str): Name of the example design.

    Returns:
        None
    """
    if project_source == "piel":
        example_design_folder = (
            os.environ["PIEL_PACKAGE_DIRECTORY"]
            + "/docs/examples/designs/"
            + example_name
        )
    elif project_source == "openlane":
        example_design_folder = (
            pathlib.Path(openlane.__file__).parent.resolve() / example_name
        )
    design_folder = os.environ["DESIGNS"] + "/" + example_name
    copy_source_folder(
        source_directory=example_design_folder, target_directory=design_folder
    )


def write_script(
    directory_path: piel_path_types,
    script: str,
    script_name: str,
) -> None:
    """
    Records a `script_name` in the `scripts` project directory.

    Args:
        directory_path(piel_path_types): Design directory.
        script(str): Script to write.
        script_name(str): Name of the script.

    Returns:
        None
    """
    directory_path = return_path(directory_path)

    directory_exists = check_path_exists(directory_path)

    if directory_exists:
        pass
    else:
        try:
            create_new_directory(directory_path)
        except PermissionError:
            print(
                UserWarning(
                    "Could not create directory: "
                    + str(directory_path.resolve())
                    + ". Your Python executable might not have the required permissions. Restructure your project directory so Python does not have to change permissions."
                )
            )

    file = open(str(directory_path / script_name), "w")
    file.write(script)
    file.close()



#!/usr/bin/env python
#
# piel documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os
import sys
import jupytext

sys.path.insert(0, os.path.abspath(".."))

import piel

master_doc = "index"  # The master toctree document.s
# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
add_module_names = False  # Remove namespaces from class/method signatures
autoapi_add_toctree_entry = True
autoapi_dirs = ["../piel"]
autoapi_keep_files = True
autoapi_template_dir = "_autoapi_templates"
autosummary_generate = True  # Turn on sphinx.ext.autosummary
autoclass_content = "both"  # Add __init__ doc (ie. params) to class summaries
autodoc_inherit_docstrings = True  # If no docstring, inherit from base class
author = "Dario Quintero"
copyright = "2023, Dario Quintero"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "examples/designs"]
extensions = [
    "autoapi.extension",
    "IPython.sphinxext.ipython_console_highlighting",
    "nbsphinx",  # Integrate Jupyter Notebooks and Sphinx
    "myst_parser",
    "sphinx.ext.autodoc",  # Core Sphinx library for auto html doc generation from docstrings
    "sphinx.ext.autosummary",  # Create neat summary tables for modules/classes/methods etc
    "sphinx.ext.coverage",
    "sphinx.ext.intersphinx",  # Link to other project's documentation (see mapping below)
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",  # Add a link to the Python source code for classes, functions etc.
    "sphinx_autodoc_typehints",  # Automatically document param types (less noise in class signature)
    "sphinx_gallery.load_style",
    "sphinx-pydantic",
]

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, "piel.tex", "piel Documentation", "Dario Quintero", "manual"),
]
html_logo = "_static/img/logo.png"
html_show_sourcelink = (
    True  # Remove 'view source code' from top of page (for html, not python)
)
html_static_path = ["_static"]
html_theme = "sphinx_book_theme"
html_title = "piel"
html_theme_options = {
    "repository_url": "https://github.com/daquintero/piel",
    "use_repository_button": True,
    "use_issues_button": True,
}
htmlhelp_basename = "pieldoc"
man_pages = [(master_doc, "piel", "piel Documentation", [author], 1)]
project = "piel"
pygments_style = "sphinx"
release = piel.__version__
set_type_checking_flag = True  # Enable 'expensive' imports for sphinx_autodoc_typehints
source_suffix = [".rst", ".md"]
templates_path = ["_templates"]
texinfo_documents = [
    (
        master_doc,
        "piel",
        "piel Documentation",
        author,
        "piel",
        "One line description of project.",
        "Miscellaneous",
    ),
]
todo_include_todos = False

mathjax3_config = {
    "tex": {"tags": "ams", "useLabelIds": True},
}
nbsphinx_allow_errors = True  # Continue through Jupyter errors
nbsphinx_custom_formats = {
    ".py": lambda s: jupytext.reads(s, ".py"),
}
nbsphinx_execute = "never"
version = piel.__version__

latex_elements: dict = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}



import gdsfactory as gf
import functools
import sax

# Defines the resistance parameters
our_resistor = functools.partial(gf.components.straight_heater_metal, ohms_per_square=2)

our_resistive_heated_waveguide = functools.partial(
    gf.components.straight_heater_metal, ohms_per_square=2
)

# our_resistive_mzi_2x2_2x2_phase_shifter = mzi2x2_2x2_phase_shifter(
#     straight_x_top=our_resistive_heated_waveguide,
# )

our_electrical_gdsfactory_netlist = our_resistive_heated_waveguide().get_netlist(
    exclude_port_types="optical"
)

"""One of the main complexities of integrating the SPICE netlist is determining the exact instance connectivity. One
of the main complexities is that in `GDSFactory` many of these electrical components are provided in the form of
cross-sectional paths. This means that we need to find a way to extract connectivity that can be translated from
these fundamental structures.

The first step is that all `GDSFactory` components need to have electrical pins attached to them too, or they can not
be extracted accordingly. It would be ideal for this to be performed as part of a subroutine in the layout generation
scheme, like for the photonic components.

This means we have to add ports onto the cross-sections for particular instantiations of these sections.
 """
# straight().show()
# print(straight().get_netlist())
# print(sax.get_required_circuit_models(straight().get_netlist()))
# We can see the opticsal ports in layers 1/10

our_resistive_heated_waveguide().show()
# print(our_resistive_heated_waveguide())
print(our_electrical_gdsfactory_netlist["instances"].keys())
print(sax.get_required_circuit_models(our_electrical_gdsfactory_netlist))
# # ['via_stack'] # Currently this is being returned, but this is not an accurate circuit model.

# We can only see optical ports, and one set of electrical ports, but this is not the full connectivity.
# This means that we need to add the ports from the path.

"""See this basic example in the `gdsfactory` `straight` definition. This means we need to do this for all cross
sections however they are implemented. All instantiations of the cross-section use this straight component definition
so we need to include in some form electrical connectivity accordingly.

    p = gf.path.straight(length=length, npoints=npoints)
    x = gf.get_cross_section(cross_section, **kwargs)

    c = Component()
    path = p.extrude(x)
    ref = c << path
    c.add_ports(ref.ports)

Note that when you extract the netlist from the straight() component, it is a fundamental component so that you cannot create a circuit out of it.
{'connections': {}, 'instances': {}, 'placements': {}, 'ports': {}, 'name': 'straight'}

We can test with an electrical port:
"""
# straight(cross_section="metal1").show()
# print(straight(cross_section="metal1").get_netlist())
"""This does not add any pins to the electrical component. Note that we need to have a function to add them in order
to extract our electrical netlist connectivity. It would be reasonable, given that the electrical connectivity can be
fanned out, that the whole metal path has pins since geometrical connectivity can be performed from all directions
without DC loss. Now, on the RF side this is a different game, but for the sake of simple connectivity this can be
done. Note that we still have to apply our device parameters in between. The problem becomes defining an input or
output, because there is none, there is only both for fundamental physical connectivity.

Let's first explore how the standard implementation fares. In order to extract physical electrical connectivity,
then we need to make sure the pin connectivity is included in between metal layers, and that VIAs allow for
connectivity between layers. This needs to conform to the standard in tools such as Cadence. In this tool,
the pin process is a different purpose than the drawing layer. I can see in the generic PDK the WG_PIN layer it was already set to purpose 10 so maybe stick with it for the standard. """
# print(gdsfactory.pdk.get_layer_stack()["core"])

# print(sax.get_required_circuit_models(straight(cross_section="metal1").get_netlist()))


@gf.cell
def connected_metal():
    test = gf.Component()
    a = test << gf.components.straight(cross_section="metal1")
    test.add_ports(a.ports)
    return test


# connected_metal().show()
# print(connected_metal().get_netlist())
# print(sax.get_required_circuit_models(connected_metal().get_netlist()))
"""
This returns this:
{'connections': {}, 'instances': {'straight_1': {'component': 'straight', 'info': {'length': 10.0, 'width': 10.0, 'cross_section': 'metal1', 'settings':
 {'width': 10.0, 'offset': 0, 'layer': 'M1', 'width_wide': None, 'auto_widen': False, 'auto_widen_minimum_length': 200.0, 'taper_length': 10.0, 'radius'
: None, 'sections': None, 'port_names': ['e1', 'e2'], 'port_types': ['electrical', 'electrical'], 'gap': 5, 'min_length': 5, 'start_straight_length': 0.
01, 'end_straight_length': 0.01, 'snap_to_grid': None, 'bbox_layers': None, 'bbox_offsets': None, 'cladding_layers': None, 'cladding_offsets': None, 'cl
adding_simplify': None, 'info': None, 'decorator': None, 'add_pins': {'function': 'add_pins', 'settings': {'function': {'function': 'add_pin_rectangle_i
nside', 'settings': {'pin_length': 0.001, 'layer_label': None}}, 'layer': 'M1_PIN'}}, 'add_bbox': None, 'mirror': False, 'name': None}, 'function_name':
 'cross_section'}, 'settings': {'cross_section': 'metal1'}}}, 'placements': {'straight_1': {'x': 0.0, 'y': 0.0, 'rotation': 0, 'mirror': 0}}, 'ports': {
'e1': 'straight_1,e1', 'e2': 'straight_1,e2'}, 'name': 'connected_metal'}
['straight']

Now we're talking business as we can create models out of this.
"""


"""
So we need to add pins at different stages including optical and electronic.

Note that part of the complexity is that the heater connectivity doesn't make sense as it is right now.
"""



import qutip  # NOQA : F401
import sax
from ..tools.qutip.unitary import matrix_to_qutip_qobj
from piel.tools.sax.utils import sax_to_s_parameters_standard_matrix

__all__ = [
    "sax_to_ideal_qutip_unitary",
    "verify_sax_model_is_unitary",
]


def sax_to_ideal_qutip_unitary(
    sax_input: sax.SType, input_ports_order: tuple | None = None
):
    """
    This function converts the calculated S-parameters into a standard Unitary matrix topology so that the shape and
    dimensions of the matrix can be observed.

    I think this means we need to transpose the output of the filtered sax SDense matrix to map it to a QuTip matrix.
    Note that the documentation and formatting of the standard `sax` mapping to a S-parameter standard notation is
    already in described in piel/piel/sax/utils.py.

    From this stage we can implement a ``QObj`` matrix accordingly and perform simulations accordingly.
    https://qutip.org/docs/latest/guide/qip/qip-basics.html#unitaries

    For example, a ``qutip`` representation of an s-gate gate would be:

    ..code-block::

        import numpy as np
        import qutip
        # S-Gate
        s_gate_matrix = np.array([[1.,   0], [0., 1.j]])
        s_gate = qutip.Qobj(mat, dims=[[2], [2]])

    In mathematical notation, this S-gate would be written as:

    ..math::

        S = \\begin{bmatrix}
            1 & 0 \\\\
            0 & i \\\\
        \\end{bmatrix}

    Args:
        sax_input (sax.SType): A dictionary of S-parameters in the form of a SDict from `sax`.
        input_ports_order (tuple | None): The order of the input ports. If None, the default order is used.

    Returns:
        qobj_unitary (qutip.Qobj): A QuTip QObj representation of the S-parameters in a unitary matrix.
    """
    # TODO make a function any SAX input.
    (
        s_parameters_standard_matrix,
        input_ports_index_tuple_order,
    ) = sax_to_s_parameters_standard_matrix(
        sax_input=sax_input, input_ports_order=input_ports_order
    )
    qobj_unitary = matrix_to_qutip_qobj(s_parameters_standard_matrix)
    return qobj_unitary


def verify_sax_model_is_unitary(
    model: sax.SType, input_ports_order: tuple | None = None
) -> bool:
    """
    Verify that the model is unitary.

    Args:
        model (dict): The model to verify.
        input_ports_order (tuple | None): The order of the input ports. If None, the default order is used.

    Returns:
        bool: True if the model is unitary, False otherwise.
    """
    qobj = sax_to_ideal_qutip_unitary(model, input_ports_order=input_ports_order)
    return qobj.check_isunitary()



import sax
from ..tools.sax import sax_to_s_parameters_standard_matrix
from ..tools.thewalrus import unitary_permanent


__all__ = ["sax_circuit_permanent", "unitary_permanent"]


def sax_circuit_permanent(
    sax_input: sax.SType,
) -> tuple:
    """
    The permanent of a unitary is used to determine the state probability of combinatorial Gaussian boson samping systems.

    ``thewalrus`` Ryser's algorithm permananet implementation is described here: https://the-walrus.readthedocs.io/en/latest/gallery/permanent_tutorial.html

    # TODO maybe implement subroutine if computation is taking forever.

    Args:
        sax_input (sax.SType): The sax S-parameter dictionary.

    Returns:
        tuple: The circuit permanent and the time it took to compute it.

    """
    s_parameter_standard_matrix, port_order = sax_to_s_parameters_standard_matrix(
        sax_input
    )
    circuit_permanent, computed_time = unitary_permanent(s_parameter_standard_matrix)
    return circuit_permanent, computed_time



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
import sax
import jax.numpy as jnp

__all__ = [
    "ideal_active_waveguide",
    "waveguide",
    "simple_straight",
    "lossless_straight",
]


def waveguide(wl=1.55, wl0=1.55, neff=2.34, ng=3.4, length=10.0, loss=0.0):
    dwl = wl - wl0
    dneff_dwl = (ng - neff) / wl0
    neff = neff - dwl * dneff_dwl
    phase = 2 * jnp.pi * neff * length / wl
    amplitude = jnp.asarray(10 ** (-loss * length / 20), dtype=complex)
    transmission = amplitude * jnp.exp(1j * phase)
    sdict = sax.reciprocal({("o1", "o2"): transmission})
    return sdict


def ideal_active_waveguide(
    wl=1.55, wl0=1.55, neff=2.34, ng=3.4, length=10.0, loss=0.0, active_phase_rad=0.0
):
    dwl = wl - wl0
    dneff_dwl = (ng - neff) / wl0
    neff = neff - dwl * dneff_dwl
    phase = (2 * jnp.pi * neff * length / wl) + active_phase_rad
    amplitude = jnp.asarray(10 ** (-loss * length / 20), dtype=complex)
    transmission = amplitude * jnp.exp(1j * phase)
    sdict = sax.reciprocal({("o1", "o2"): transmission})
    return sdict


def simple_straight(length=10.0, width=0.5):
    S = {("o1", "o2"): 1.0}  # we'll improve this model later!
    return sax.reciprocal(S)


def lossless_straight():
    S = {("o1", "o2"): 1.0}  # we'll improve this model later!
    return sax.reciprocal(S)



from ....file_system import return_path, get_files_recursively_in_directory

__all__ = [
    "filter_timing_sta_files",
    "filter_power_sta_files",
    "get_all_timing_sta_files",
    "get_all_power_sta_files",
]


def filter_timing_sta_files(file_list):
    """
    Filter the timing sta files from the list of files

    Args:
        file_list (list): List containing the file paths

    Returns:
        timing_sta_files (list): List containing the timing sta files
    """
    timing_sta_files = []
    for file_path in file_list:
        if file_path.endswith(("sta.rpt", "sta.min.rpt", "sta.max.rpt")):
            timing_sta_files.append(file_path)
    return timing_sta_files


def filter_power_sta_files(file_list):
    """
    Filter the power sta files from the list of files

    Args:
        file_list (list): List containing the file paths

    Returns:
        power_sta_files (list): List containing the power sta files
    """
    power_sta_files = []
    for file_path in file_list:
        if file_path.endswith(("power.rpt")):
            power_sta_files.append(file_path)
    return power_sta_files


def get_all_timing_sta_files(run_directory):
    """
    This function aims to list and perform analysis on all the relevant files in a particular run between all the corners.

    Args:
        run_directory (str, optional): The run directory to perform the analysis on. Defaults to None.

    Returns:
        timing_sta_files_list (list): List of all the .rpt files in the run directory.
    """
    run_directory = return_path(run_directory)
    all_rpt_files_list = get_files_recursively_in_directory(run_directory, "rpt")
    timing_sta_files_list = filter_timing_sta_files(all_rpt_files_list)
    return timing_sta_files_list


def get_all_power_sta_files(run_directory):
    """
    This function aims to list and perform analysis on all the relevant files in a particular run between all the corners.

    Args:
        run_directory (str, optional): The run directory to perform the analysis on. Defaults to None.

    Returns:
        power_sta_files_list (list): List of all the .rpt files in the run directory.
    """
    run_directory = return_path(run_directory)
    all_rpt_files_list = get_files_recursively_in_directory(run_directory, "rpt")
    power_sta_files_list = filter_power_sta_files(all_rpt_files_list)
    return power_sta_files_list



"""
In this function we implement different methods of mapping electronic signals to phase.

One particular implementation of phase mapping would be:

.. list-table:: Example Basic Phase Mapping
   :header-rows: 1

   * - Bit
     - Phase
   * - b0
     - :math:`\\phi_0 \\to 0`
   * - b1
     - :math:`\\phi_1 \\to \\pi`

We can define the two corresponding angles that this would be.

A more complex implementation of phase mapping can be similar to a DAC mapping: a bitstring within a converter bit-size can map directly to a particular phase space within a particular mapping.
"""
import numpy as np
import pandas as pd

__all__ = [
    "bits_array_from_bits_amount",
    "linear_bit_phase_map",
    "return_phase_array_from_data_series",
]


def bits_array_from_bits_amount(bits_amount: int) -> np.ndarray:
    """
    Returns an array of bits from a given amount of bits.

    Args:
        bits_amount(int): Amount of bits to generate.

    Returns:
        bit_array(np.ndarray): Array of bits.
    """
    maximum_integer_represented = 2 ** (bits_amount)
    int_array = np.arange(maximum_integer_represented)
    bit_array = np.vectorize(np.base_repr)(int_array)
    return bit_array


def linear_bit_phase_map(
    bits_amount: int,
    final_phase_rad: float,
    initial_phase_rad: float = 0,
    return_dataframe: bool = True,
    quantization_error: float = 0.000001,
) -> dict | pd.DataFrame:
    """
    Returns a linear direct mapping of bits to phase.

    Args:
        bits_amount(int): Amount of bits to generate.
        final_phase_rad(float): Final phase to map to.
        initial_phase_rad(float): Initial phase to map to.

    Returns:
        bit_phase_mapping(dict): Mapping of bits to phase.
    """
    bits_array = bits_array_from_bits_amount(bits_amount)
    phase_division_amount = len(bits_array) - 1
    phase_division_step = (
        final_phase_rad - initial_phase_rad
    ) / phase_division_amount - quantization_error
    linear_phase_array = np.arange(
        initial_phase_rad, final_phase_rad, phase_division_step
    )
    bit_phase_mapping_raw = {
        "bits": bits_array,
        "phase": linear_phase_array,
    }
    if return_dataframe:
        bit_phase_mapping = pd.DataFrame(bit_phase_mapping_raw)
    else:
        bit_phase_mapping = bit_phase_mapping_raw
    return bit_phase_mapping


def return_phase_array_from_data_series(
    data_series: pd.Series,
    phase_map: pd.DataFrame | pd.Series,
) -> list:
    """
    Returns a list of phases from a given data series and phase map.
    # TODO optimise lookup table speed

    Args:
        data_series(pd.Series): Data series to map.
        phase_map(pd.DataFrame | pd.Series): Phase map to use.

    Returns:
        phase_array(list): List of phases.
    """
    phase_array = []
    for code_i in data_series.values:
        phase = phase_map[phase_map.bits == str(code_i)].phase.values[0]
        phase_array.append(phase)
    return phase_array



"""
The objective of this file is to provide the simulation ports and interconnection to consider modelling digital and mixed signal logic.

The main simulation driver is cocotb, and this generates a set of files that correspond to time-domain digital
simulations. The cocotb verification software can also be used to perform mixed signal simulation, and digital data
can be inputted as a bitstream into a photonic solver, although the ideal situation would be to have integrated
photonic time-domain models alongside the electronic simulation solver, and maybe this is where it will go. It can be
assumed that, as is currently, cocotb can interface python with multiple solvers until someone (and I'd love to do
this) writes an equivalent python-based or C++ based python time-domain simulation solver.

The nice thing about cocotb is that as long as the photonic simulations can be written asynchronously, time-domain
simulations can be closely integrated or simulated through this verification software. """
import functools
import pathlib
import subprocess
from typing import Literal
from piel.file_system import return_path, write_script, delete_path_list_in_directory

__all__ = [
    "check_cocotb_testbench_exists",
    "configure_cocotb_simulation",
    "delete_simulation_output_files",
    "run_cocotb_simulation",
]


def check_cocotb_testbench_exists(
    design_directory: str | pathlib.Path,
) -> bool:
    """
    Checks if a cocotb testbench exists in the design directory.

    Args:
        design_directory(str | pathlib.Path): Design directory.

    Returns:
        cocotb_testbench_exists(bool): True if cocotb testbench exists.
    """
    cocotb_testbench_exists = False
    design_directory = return_path(design_directory)
    testbench_directory = design_directory / "tb"
    testbench_directory_exists = testbench_directory.exists()

    if testbench_directory_exists:
        # Check if cocotb python files are present
        cocotb_python_files = list(testbench_directory.glob("*.py"))
        if len(cocotb_python_files) > 0:
            cocotb_testbench_exists = True
        else:
            pass
    else:
        pass

    return cocotb_testbench_exists


def configure_cocotb_simulation(
    design_directory: str | pathlib.Path,
    simulator: Literal["icarus", "verilator"],
    top_level_language: Literal["verilog", "vhdl"],
    top_level_verilog_module: str,
    test_python_module: str,
    design_sources_list: list | None = None,
):
    """
    Writes a cocotb makefile.

    If no design_sources_list is provided then it adds all the design sources under the `src` folder.

    In the form
    .. code-block::

        #!/bin/sh
        # Makefile
        # defaults
        SIM ?= icarus
        TOPLEVEL_LANG ?= verilog

        # Note we need to include the test script to the PYTHONPATH
        export PYTHONPATH =

        VERILOG_SOURCES += $(PWD)/my_design.sv
        # use VHDL_SOURCES for VHDL files

        # TOPLEVEL is the name of the toplevel module in your Verilog or VHDL file
        TOPLEVEL := my_design

        # MODULE is the basename of the Python test file
        MODULE := test_my_design

        # include cocotb's make rules to take care of the simulator setup
        include $(shell cocotb-config --makefiles)/Makefile.sim


    Args:
        design_directory (str | pathlib.Path): The directory where the design is located.
        simulator (Literal["icarus", "verilator"]): The simulator to use.
        top_level_language (Literal["verilog", "vhdl"]): The top level language.
        top_level_verilog_module (str): The top level verilog module.
        test_python_module (str): The test python module.
        design_sources_list (list | None, optional): A list of design sources. Defaults to None.

    Returns:
        None
    """
    design_directory = return_path(design_directory)
    design_sources_directory = design_directory / "src"

    if design_sources_list is not None:
        # Include all the design sources files in a list
        design_sources_list = list(design_sources_directory.iterdir())

    top_commands_list = [
        "#!/bin/bash",
        "# Makefile",
        "SIM ?= " + simulator,
        "TOPLEVEL_LANG ?= " + top_level_language,
    ]

    middle_commands_list = []
    # TODO: Implement mixed source designs.
    if top_level_language == "verilog":
        for source_file in design_sources_list:
            middle_commands_list.append(
                "VERILOG_SOURCES += " + str(source_file.resolve())
            )
    elif top_level_language == "vhdl":
        for source_file in design_sources_list:
            middle_commands_list.append("VHDL_SOURCES += " + str(source_file.resolve()))

    bottom_commands_list = [
        "TOPLEVEL := " + top_level_verilog_module,
        "MODULE := " + test_python_module,
        "include $(shell cocotb-config --makefiles)/Makefile.sim",
    ]

    commands_list = []
    commands_list.extend(top_commands_list)
    commands_list.extend(middle_commands_list)
    commands_list.extend(bottom_commands_list)

    script = " \n".join(commands_list)
    write_script(
        directory_path=design_directory / "tb", script=script, script_name="Makefile"
    )


delete_simulation_output_files = functools.partial(
    delete_path_list_in_directory,
    path_list=["sim_build", "__pycache__", "ivl_vhdl_work"],
)


def run_cocotb_simulation(
    design_directory: str,
) -> subprocess.CompletedProcess:
    """
    Equivalent to running the cocotb makefile
    .. code-block::

        make

    Args:
        design_directory (str): The directory where the design is located.

    Returns:
        subprocess.CompletedProcess: The subprocess.CompletedProcess object.

    """
    test_directory = return_path(design_directory) / "tb"
    commands_list = ["cd " + str(test_directory.resolve()), "make"]
    script = "; \n".join(commands_list)
    # Save script if desired to run directly
    write_script(
        directory_path=test_directory,
        script=script,
        script_name="run_cocotb_simulation.sh",
    )
    run = subprocess.run(script, capture_output=True, shell=True, check=True)
    return run



import pandas as pd
import pathlib
from piel.file_system import return_path
from .utils import (
    contains_in_lines,
    read_file_lines,
    get_file_line_by_keyword,
    create_file_lines_dataframe,
)

__all__ = [
    "calculate_max_frame_amount",
    "calculate_propagation_delay_from_file",
    "calculate_propagation_delay_from_timing_data",
    "configure_timing_data_rows",
    "configure_frame_id",
    "filter_timing_data_by_net_name_and_type",
    "get_frame_meta_data",
    "get_frame_lines_data",
    "get_frame_timing_data",
    "get_all_timing_data_from_file",
    "read_sta_rpt_fwf_file",
]


def calculate_max_frame_amount(
    file_lines_data: pd.DataFrame,
):
    """
    Calculate the maximum frame amount based on the frame IDs in the DataFrame

    Args:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines

    Returns:
        maximum_frame_amount (int): Maximum number of frames in the file
    """
    return max(file_lines_data["frame_id"].tolist()) + 1


def calculate_propagation_delay_from_timing_data(
    net_name_in: str,
    net_name_out: str,
    timing_data: pd.DataFrame,
):
    """
    Calculate the propagation delay between two nets

    Args:
        net_name_in (str): Name of the input net
        net_name_out (str): Name of the output net
        timing_data (pd.DataFrame): Dataframe containing the timing data

    Returns:
        propagation_delay_dataframe (pd.DataFrame): Dataframe containing the propagation delay
    """
    if (
        len(filter_timing_data_by_net_name_and_type(timing_data, net_name_in, "in")) > 0
        and len(
            filter_timing_data_by_net_name_and_type(timing_data, net_name_out, "out")
        )
        > 0
    ):
        output_net_dataframe = (
            timing_data[(timing_data.net_type == "out")]
            .copy()
            .add_suffix("_out")
            .reset_index()
        )
        input_net_dataframe = (
            timing_data[(timing_data.net_type == "in")]
            .copy()
            .add_suffix("_in")
            .reset_index()
        )
        propagation_delay_dataframe = output_net_dataframe.merge(
            input_net_dataframe, left_index=True, right_index=True
        )
        propagation_delay_dataframe["propagation_delay"] = pd.to_numeric(
            output_net_dataframe.iloc[:].Time_out.values
        ) - pd.to_numeric(input_net_dataframe.iloc[:].Time_in.values)
        return propagation_delay_dataframe


def calculate_propagation_delay_from_file(
    file_path: str | pathlib.Path,
):
    """
    Calculate the propagation delay for each frame in the file

    Args:
        file_path (str | pathlib.Path): Path to the file

    Returns:
        propagation_delay (dict): Dictionary containing the propagation delay
    """
    # TODO Check file is RPT
    file_lines_data = get_frame_lines_data(file_path)
    maximum_frame_amount = calculate_max_frame_amount(file_lines_data)
    (
        start_point_name,
        end_point_name,
        path_group_name,
        path_type_name,
    ) = get_frame_meta_data(file_lines_data)

    frame_timing_data = get_all_timing_data_from_file(file_path)
    propagation_delay = {}
    for frame_id in range(maximum_frame_amount):
        if len(start_point_name.values) > frame_id:
            frame_net_in = start_point_name.values[frame_id][0]
            frame_net_out = end_point_name.values[frame_id][0]
            propagation_delay[frame_id] = calculate_propagation_delay_from_timing_data(
                frame_net_in, frame_net_out, frame_timing_data[frame_id]
            )
    return propagation_delay


def configure_timing_data_rows(
    file_lines_data: pd.DataFrame,
):
    """
    Identify the timing data lines for each frame and creates a metadata dictionary for frames.

    Args:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines

    Returns:
        frame_meta_data (dict): Dictionary containing the frame metadata
    """
    frame_meta_data = {}
    maximum_frame_amount = calculate_max_frame_amount(file_lines_data)
    for frame in range(maximum_frame_amount):
        timing_rows_index = file_lines_data.index[
            file_lines_data["timing_data_line"] & (file_lines_data["frame_id"] == frame)
        ].tolist()
        if len(timing_rows_index) > 0:
            frame_meta_data[frame] = {
                "start_index": timing_rows_index[0],
                "end_index": timing_rows_index[-1],
                "start_rows_skip": timing_rows_index[0] + 1,
                "end_rows_skip": len(file_lines_data) - timing_rows_index[-1],
            }
        else:
            frame_meta_data[frame] = {
                "start_index": 0,
                "end_index": 0,
                "start_rows_skip": 0,
                "end_rows_skip": 0,
            }
    return frame_meta_data


def configure_frame_id(
    file_lines_data: pd.DataFrame,
):
    """
    Identify the frame delimiters and assign frame ID to each line in the file

    Args:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines

    Returns:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines
    """
    file_lines_data["delimiters_line"] = file_lines_data.lines.str.contains(
        "=========="
    )
    file_lines_data["timing_data_line"] = file_lines_data.lines.str.contains(
        "---------"
    )
    frame_data = []
    frame_id_counter = 0
    parity_counter = 1
    for delimiter in file_lines_data["delimiters_line"]:
        if delimiter:
            if parity_counter % 2:
                frame_id_counter += 1
                parity_counter = 0
            else:
                parity_counter += 1
        frame_data.append(frame_id_counter - 1)
    file_lines_data["frame_id"] = frame_data
    return file_lines_data


def filter_timing_data_by_net_name_and_type(
    timing_data: pd.DataFrame, net_name: str, net_type: str
):
    """
    Filter the timing data by net name and type

    Args:
        timing_data (pd.DataFrame): DataFrame containing the timing data
        net_name (str): Net name to be filtered
        net_type (str): Net type to be filtered

    Returns:
        timing_data (pd.DataFrame): DataFrame containing the timing data
    """
    return timing_data[
        (timing_data.net_name == net_name) & (timing_data.net_type == net_type)
    ]


def get_frame_meta_data(file_lines_data):
    """
    Get the frame metadata

    Args:
        file_lines_data (pd.DataFrame): DataFrame containing the file lines

    Returns:
        start_point_name (pd.DataFrame): DataFrame containing the start point name
        end_point_name (pd.DataFrame): DataFrame containing the end point name
        path_group_name (pd.DataFrame): DataFrame containing the path group name
        path_type_name (pd.DataFrame): DataFrame containing the path type name
    """
    file_lines_data["start_point_line"] = contains_in_lines(
        file_lines_data, "Startpoint"
    )
    file_lines_data["end_point_line"] = contains_in_lines(file_lines_data, "Endpoint")
    file_lines_data["path_group_line"] = contains_in_lines(
        file_lines_data, "Path Group"
    )
    file_lines_data["path_type_line"] = contains_in_lines(file_lines_data, "Path Type")

    start_point_name = get_file_line_by_keyword(
        file_lines_data, "start_point", r"((?<=Startpoint:\s).*?(?=\s\())"
    )
    end_point_name = get_file_line_by_keyword(
        file_lines_data, "end_point", r"((?<=Endpoint:\s).*?(?=\s\())"
    )
    path_group_name = get_file_line_by_keyword(
        file_lines_data, "path_group", r"((?<=Path Group:\s).*)"
    )
    path_type_name = get_file_line_by_keyword(
        file_lines_data, "path_type", r"((?<=Path Type:\s).*)"
    )

    return start_point_name, end_point_name, path_group_name, path_type_name


def get_frame_lines_data(
    file_path: str | pathlib.Path,
):
    """
    Calculate the timing data for each frame in the file

    Args:
        file_path (str | pathlib.Path): Path to the file

    Returns:
        file_lines_data (pd.DataFrame): DataFrame containing the file lines
    """
    file_path = return_path(file_path)
    # TODO Check file is RPT
    file_lines = read_file_lines(file_path)
    file_lines_data = create_file_lines_dataframe(file_lines)
    file_lines_data = configure_frame_id(file_lines_data)
    return file_lines_data


def get_frame_timing_data(
    file: str | pathlib.Path, frame_meta_data: dict, frame_id: int = 0
):
    """
    Extract the timing data from the file

    Args:
        file (str | pathlib.Path): Address of the file
        frame_meta_data (dict): Dictionary containing the frame metadata
        frame_id (int): Frame ID to be read

    Returns:
        timing_data (pd.DataFrame): DataFrame containing the timing data
    """
    file = return_path(file)
    timing_data = read_sta_rpt_fwf_file(file, frame_meta_data, frame_id)
    timing_data["net_type"] = timing_data["Description"].str.extract(r"\(([^()]+)\)")
    timing_data["net_name"] = timing_data["Description"].str.extract(r"(.*?)\s?\(.*?\)")
    return timing_data


def get_all_timing_data_from_file(
    file_path: str | pathlib.Path,
):
    """
    Calculate the timing data for each frame in the file

    Args:
        file_path (str | pathlib.Path): Path to the file

    Returns:
        frame_timing_data (dict): Dictionary containing the timing data for each frame
    """
    file_lines_data = get_frame_lines_data(file_path)
    maximum_frame_amount = calculate_max_frame_amount(file_lines_data)
    frame_meta_data = configure_timing_data_rows(file_lines_data)
    frame_timing_data = {}
    for frame_id in range(maximum_frame_amount):
        frame_timing_data[frame_id] = get_frame_timing_data(
            file_path, frame_meta_data, frame_id
        )
    return frame_timing_data


def read_sta_rpt_fwf_file(
    file: str | pathlib.Path,
    frame_meta_data: dict,
    frame_id: int = 0,
):
    """
    Read the fixed width file and return a DataFrame

    Args:
        file (str | pathlib.Path): Address of the file
        frame_meta_data (dict): Dictionary containing the frame metadata
        frame_id (int): Frame ID to be read

    Returns:
        file_data (pd.DataFrame): DataFrame containing the file data
    """
    file = return_path(file)
    file_data = pd.read_fwf(
        str(file.resolve()),
        colspecs=[(0, 6), (6, 14), (14, 22), (22, 30), (30, 38), (38, 40), (40, 100)],
        skiprows=frame_meta_data[frame_id]["start_rows_skip"],
        skipfooter=frame_meta_data[frame_id]["end_rows_skip"],
        names=["Fanout", "Cap", "Slew", "Delay", "Time", "Direction", "Description"],
    )
    return file_data



"""
`sax` has very good GDSFactory integration functions, so there is a question on whether implementing our own circuit
construction, and SPICE netlist parser from it, accordingly. We need in some form to connect electrical models to our
parsed netlist, in order to apply SPICE passive values, and create connectivity for each particular device. Ideally,
this would be done from the component instance as that way the component model can be integrated with its geometrical
parameters, but does not have to be done necessarily. This comes down to implementing a backend function to compile
SAX compiled circuit.
"""
import copy
import networkx as nx
from sax.circuit import (
    create_dag,
    _ensure_recursive_netlist_dict,
    remove_unused_instances,
    _extract_instance_models,
    _validate_net,
    _validate_dag,
    _validate_models,
)
from sax.netlist import RecursiveNetlist

from ...models.physical.electronic import get_default_models
from .utils import convert_tuples_to_strings

__all__ = [
    "convert_connections_to_tuples",
    "gdsfactory_netlist_with_hdl21_generators",
]


def convert_connections_to_tuples(connections: dict):
    """
    Convert from:

    .. code-block::

        {
        'straight_1,e1': 'taper_1,e2',
        'straight_1,e2': 'taper_2,e2',
        'taper_1,e1': 'via_stack_1,e3',
        'taper_2,e1': 'via_stack_2,e1'
        }

    to:

    .. code-block::

        [(('straight_1', 'e1'), ('taper_1', 'e2')), (('straight_1', 'e2'), ('taper_2', 'e2')), (('taper_1', 'e1'),
        ('via_stack_1', 'e3')), (('taper_2', 'e1'), ('via_stack_2', 'e1'))]
    """
    tuples_list = []
    for key, value in connections.items():
        source_tuple = tuple(key.split(","))
        dest_tuple = tuple(value.split(","))
        tuples_list.append((source_tuple, dest_tuple))
    return tuples_list


def get_matching_connections(names: list, connections: dict):
    """
    This function returns a list of tuples that match the names of the connections.

    Args:
        names: List of names to match
        connections: Dictionary of connections to match

    Returns:
        List of tuples that match the names of the connections
    """
    matching_connections = []
    for key, value in connections.items():
        for name in names:
            if name in key or name in value:
                matching_connections.append((key, value))
    return matching_connections


def get_matching_port_nets(names, connections):
    """
    This function returns a list of tuples that match the names of the connections.

    Args:
        names: List of names to match
        connections: Dictionary of connections to match

    Returns:
        List of tuples that match the names of the connections
    """
    matching_tuples = get_matching_connections(names, connections)
    matching_strings = convert_tuples_to_strings(matching_tuples)
    return matching_strings


def gdsfactory_netlist_with_hdl21_generators(
    gdsfactory_netlist: dict,
    generators=None,
):
    """
    This function allows us to map the ``hdl21`` models dictionary in a `sax`-like implementation to the ``GDSFactory`` netlist. This allows us to iterate over each instance in the netlist and construct a circuit after this function.]

    Example usage:

    .. code-block::

        >>> import gdsfactory as gf
        >>> from piel.integration.gdsfactory_hdl21.conversion import gdsfactory_netlist_with_hdl21_generators
        >>> from piel.models.physical.electronic import get_default_models
        >>> gdsfactory_netlist_with_hdl21_generators(gdsfactory_netlist=gf.components.mzi2x2_2x2_phase_shifter().get_netlist(exclude_port_types="optical"),generators=get_default_models())

    Args:
        gdsfactory_netlist: The netlist from ``GDSFactory`` to map to the ``hdl21`` models dictionary.
        generators: The ``hdl21`` models dictionary to map to the ``GDSFactory`` netlist.

    Returns:
        The ``GDSFactory`` netlist with the ``hdl21`` models dictionary.
    """
    electrical_models_netlist = copy.copy(gdsfactory_netlist)
    if generators is None:
        generators = get_default_models()

    netlist = _ensure_recursive_netlist_dict(gdsfactory_netlist)

    # TODO: do the following two steps *after* recursive netlist parsing.
    netlist = remove_unused_instances(netlist)
    netlist, instance_models = _extract_instance_models(netlist)

    recnet: RecursiveNetlist = _validate_net(netlist)
    dependency_dag: nx.DiGraph = _validate_dag(
        create_dag(recnet, generators)
    )  # directed acyclic graph
    generators = _validate_models(
        {**(generators or {}), **instance_models}, dependency_dag
    )

    new_models = {}
    current_models = {}
    model_names = list(nx.topological_sort(dependency_dag))[::-1]
    for model_name in model_names:
        if model_name in generators:
            new_models[model_name] = generators[model_name]
            continue

        flatnet = recnet.__root__[model_name]
        current_models.update(new_models)
        new_models = {}
        inst2model = {
            k: generators[inst.component] for k, inst in flatnet.instances.items()
        }

        # Iterate over every instance and append the corresponding required SPICE connectivity
        for instance_name_i, _ in list(electrical_models_netlist["instances"].items()):
            electrical_models_netlist["instances"][instance_name_i][
                "hdl21_model"
            ] = inst2model[instance_name_i]
    return electrical_models_netlist


# Not for the `hdl21` design flow.
def gdsfactory_netlist_to_spice_string_connectivity_netlist(
    gdsfactory_netlist: dict,
    models=None,
):
    """
    Not for the `hdl21` design flow, but useful for other SPICE level conversions.

    This function maps the connections of a netlist to a node that can be used in a SPICE netlist. SPICE netlists are
    in the form of:

    .. code-block:: spice

        RXXXXXXX N1 N2 <VALUE> <MNAME> <L=LENGTH> <W=WIDTH> <TEMP=T>

    This means that every instance, is an electrical type, and we define the two particular nodes in which it is
    connected. This means we need to convert the gdsfactory dictionary netlist into a form that allows us to map the
    connectivity for every instance. Then we can define that as a line of the SPICE netlist with a particular
    electrical model. For passives this works fine when it's a two port network such as sources, or electrical
    elements. However, non-passive elements like transistors have three ports or more which are provided in an ordered form.

    This means that the order of translations is as follows:

    .. code-block::

        1. Extract all instances and required models from the netlist, and assign the corresponding parameters on instantiation.
        2. Verify that the models have been provided. Each model describes the type of component this is, how many ports it requires and so on.
        3. Map the connections to each instance port as part of the instance dictionary.

    We should get as an output a dictionary in the structure:

    .. code-block::

        {
            instance_1: {
                ...
                "connections": [('straight_1,e1', 'taper_1,e2'),
                                ('straight_1,e2', 'taper_2,e2')],
                'spice_nets': {'e1': 'straight_1__e1___taper_1__e2',
                        'e2': 'straight_1__e2___taper_2__e2'},
                'spice_model': <function piel.models.physical.electronic.spice.resistor.basic_resistor()>},
            }
            ...
        }
    """
    spice_netlist = copy.copy(gdsfactory_netlist)
    if models is None:
        models = get_default_models()

    netlist = _ensure_recursive_netlist_dict(gdsfactory_netlist)

    # TODO: do the following two steps *after* recursive netlist parsing.
    netlist = remove_unused_instances(netlist)
    netlist, instance_models = _extract_instance_models(netlist)

    recnet: RecursiveNetlist = _validate_net(netlist)
    dependency_dag: nx.DiGraph = _validate_dag(
        create_dag(recnet, models)
    )  # directed acyclic graph
    models = _validate_models({**(models or {}), **instance_models}, dependency_dag)

    new_models = {}
    current_models = {}
    model_names = list(nx.topological_sort(dependency_dag))[::-1]
    for model_name in model_names:
        if model_name in models:
            new_models[model_name] = models[model_name]
            continue

        flatnet = recnet.__root__[model_name]
        current_models.update(new_models)
        new_models = {}
        inst2model = {
            k: models[inst.component] for k, inst in flatnet.instances.items()
        }

        # Iterate over every instance and append the corresponding required SPICE connectivity
        for instance_name_i, _ in list(spice_netlist["instances"].items()):
            spice_netlist["instances"][instance_name_i][
                "connections"
            ] = get_matching_connections(
                names=[instance_name_i], connections=gdsfactory_netlist["connections"]
            )
            spice_netlist["instances"][instance_name_i][
                "spice_nets"
            ] = get_matching_port_nets(
                names=[instance_name_i], connections=gdsfactory_netlist["connections"]
            )
            spice_netlist["instances"][instance_name_i]["spice_model"] = inst2model[
                instance_name_i
            ]
    return spice_netlist



test_spm_open_lane_configuration = {
    "PDK": "sky130A",
    "DESIGN_NAME": "spm",
    "VERILOG_FILES": ["./src/spm.v"],
    "CLOCK_PORT": "clk",
    "CLOCK_PERIOD": 10,
}

example_open_lane_configuration = {
    "CLOCK_PERIOD": 100,
    "CLOCK_NET": "clk_o",
    "CLOCK_PORT": None,
    "CLOCK_TREE_SYNTH": True,
    "DIE_AREA": "0 0 100.0 100.00",
    "DESIGN_NAME": None,
    "FP_CORE_UTIL": 40,
    "FP_PDN_AUTO_ADJUST": 0,
    "FP_PDN_VPITCH": 25,
    "FP_PDN_HPITCH": 25,
    "FP_PDN_VOFFSET": 5,
    "FP_PDN_HOFFSET": 5,
    "FP_PIN_ORDER_CFG": "dir::pin_order.cfg",
    "FP_SIZING": "absolute",
    "PL_TARGET_DENSITY": 0.75,
    "PL_RANDOM_GLB_PLACEMENT": True,
    "SYNTH_NO_FLAT": 1,
    "SYNTH_CAP_LOAD": 100,
    "SYNTH_FLAT_TOP": True,
    "SYNTH_MAX_FANOUT": 10,
    "SYNTH_PARAMETERS": "",
    "VERILOG_FILES": None,
    "VERILOG_INCLUDE_DIRS": "dir::src/",
    "pdk::sky130*": {
        "FP_CORE_UTIL": 45,
        "scl::sky130_fd_sc_hd": {"CLOCK_PERIOD": 10},
        "scl::sky130_fd_sc_hdll": {"CLOCK_PERIOD": 10},
        "scl::sky130_fd_sc_hs": {"CLOCK_PERIOD": 8},
        "scl::sky130_fd_sc_ls": {"CLOCK_PERIOD": 10, "SYNTH_MAX_FANOUT": 5},
        "scl::sky130_fd_sc_ms": {"CLOCK_PERIOD": 10},
    },
    "pdk::gf180mcu*": {
        "CLOCK_PERIOD": 24.0,
        "FP_CORE_UTIL": 40,
        "SYNTH_MAX_FANOUT": 4,
        "PL_TARGET_DENSITY": 0.5,
    },
}



__all__ = ["convert_awg_to_m2"]


def convert_awg_to_m2(awg: int) -> float:
    """
    Converts an AWG value to meters squared.

    Reference: https://en.wikipedia.org/wiki/American_wire_gauge

    Args:
        awg (ing): AWG value.

    Returns:
        float: Cross sectional area in meters squared.
    """
    return ((0.127) * (92 ** ((36 - awg) / 39))) * 1e-3



import numpy as np
import pandas as pd

__all__ = [
    "append_row_to_dict",
    "points_to_lines_fixed_transient",
]


def append_row_to_dict(
    data: dict,
    copy_index: int,
    set_value: dict,
):
    """
    Get all the rows of the dictionary. We want to copy and append a row at a particular index of the dictionary values.
    Operates on existing data

    Args:
        data: Dictionary of data to be appended.
        copy_index: Index of the row to be copied.
        set_value: Dictionary of values to be set at the copied index.

    Returns:
        None
    """
    keys_list = list(data.keys())
    for key in keys_list:
        # Iterates over each key
        # Gets data at key and appends into dictionary at the end
        index_length = len(data[key])
        if type(data[key]) == list:
            data[key].append(data[key][copy_index])
        elif type(data[key]) == np.ndarray:
            data[key] = np.append(data[key], data[key][copy_index])
        elif type(data[key]) == dict:
            # Assumes a key,value {index: value} form that starts from 0
            # Find length of the dictionary
            data[key][index_length] = data[key][copy_index]
        else:
            raise ValueError(
                "data[key] invalid " + str(data[key]) + " for key: " + str(key)
            )

        if key in set_value.keys():
            # If value to set in the key set of the dictionary then update copied row latest appended
            if (type(data[key]) == list) or (type(data[key]) == np.ndarray):
                data[key][-1] = set_value[key]
            elif type(data[key]) == dict:
                data[key][index_length] = set_value[key]

    return data


def points_to_lines_fixed_transient(
    data: pd.DataFrame | dict,
    time_index_name: str,
    fixed_transient_time=1,
    return_dict: bool = False,
):
    """
    This function converts specific steady-state point data into steady-state lines with a defined transient time in order to plot digital-style data.

    For example, VCD data tends to be structured in this form:

    .. code-block:: text

        #2001
        b1001 "
        b10010 #
        b1001 !
        #4001
        b1011 "
        b1011 #
        b0 !
        #6001
        b101 "

    This means that even when tokenizing the data, when visualising it in a wave plotter such as GTKWave, the signals
    get converted from token specific times to transient signals by a corresponding transient rise time. If we want
    to plot the data correspondingly in Python, it is necessary to add some form of transient signal translation.
    Note that this operates on a dataframe where the electrical time signals are clearly defined. It copies the
    corresponding steady-state data points whilst adding data points for the time-index accordingly.

    It starts by creating a copy of the initial dataframe as to not overwrite the existing data. We have an initial
    time data point that tends to start at time 0. This means we need to add a point just before the next steady
    state point transition. So what we want to do is copy the existing row and just change the time to be the
    `fixed_transient_time` before the next transition.

    Doesn't append on penultimate row.

    Args:
        data: Dataframe or dictionary of data to be converted.
        time_index_name: Name of the time index column.
        fixed_transient_time: Time of the transient signal.
        return_dict: Return a dictionary instead of a dataframe.

    Returns:
        Dataframe or dictionary of data with steady-state lines.
    """
    if type(data) == pd.DataFrame:
        data = data.to_dict()
        data = data.copy()
    else:
        data = data.copy()

    for i in range(len(data[time_index_name]) - 1):
        # Create a copy of the first row with a i+1 index.
        if i == (len(data[time_index_name]) - 2):
            # Check if on penultimate row don't append
            pass
        else:
            new_steady_time = data[time_index_name][i + 1] - fixed_transient_time
            append_row_to_dict(
                data=data, copy_index=i, set_value={time_index_name: new_steady_time}
            )

    if return_dict:
        pass
    else:
        data = pd.DataFrame(data).sort_values(by="t")

    return data



from piel.models import frequency  # NOQA: F401
from piel.models import logic  # NOQA: F401
from piel.models import physical  # NOQA: F401
from piel.models import transient  # NOQA: F401

from .utils import *



import sax
import gdsfactory as gf


@gf.cell
def connected_metal():
    test = gf.Component()
    a = test << gf.components.straight(cross_section="metal1")
    b = test << gf.components.straight(cross_section="metal1")
    a.connect("e1", b.ports["e1"])
    test.add_port(port=a.ports["e2"], name="in")
    test.add_port(port=b.ports["e2"], name="out")
    return test


our_connected_metal = connected_metal()
# import json
# print(json.dumps(our_connected_metal.get_netlist()))
# print(our_connected_metal.get_netlist())

our_connected_metal.show()
print(sax.netlist(our_connected_metal.get_netlist()))
# print(dir(sax.netlist(connected_metal().get_netlist())))
"""
RecursiveNetlist(
    __root__={
        "top_level": Netlist(
            instances={
                "straight_1": Component(
                    component="straight", settings={"cross_section": "metal1"}
                ),
                "straight_2": Component(
                    component="straight", settings={"cross_section": "metal1"}
                ),
            },
            connections={"straight_1,e1": "straight_2,e1"},
            ports={"in": "straight_1,e2", "out": "straight_2,e2"},
            placements={
                "straight_1": Placement(
                    x="0.0",
                    y="0.0",
                    xmin=None,
                    ymin=None,
                    xmax=None,
                    ymax=None,
                    dx=0,
                    dy=0,
                    port=None,
                    rotation=180,
                    mirror=False,
                ),
                "straight_2": Placement(
                    x="0.0",
                    y="0.0",
                    xmin=None,
                    ymin=None,
                    xmax=None,
                    ymax=None,
                    dx=0,
                    dy=0,
                    port=None,
                    rotation=0,
                    mirror=False,
                ),
            },
        )
    }
)
"""



import hdl21 as h

__all__ = ["taper", "TaperParameters"]


@h.paramclass
class TaperParameters:
    pass


@h.generator
def taper(params: TaperParameters) -> h.Module:
    """
    Implements a `hdl21` taper resistor class. We need to include the mapping ports as we expect our gdsfactory component to be with the instance of the model.
    """

    @h.module
    class Taper:
        e1, e2 = h.Ports(2)
        r1 = h.IdealResistor(r=1e3)(p=e1, n=e2)

    return Taper



"""The way the construction of the SPICE models is implemented in ``piel`` is also microservice-esque. Larger
circuits are constructed out of smaller building blocks. This means that simulation configurations and so on are
constructed upon a provided initial circuit, and the SPICE directives are appended accordingly.

As mentioned previously, ``piel`` creates circuits based on fundamental SPICE because this opens the possibility to
large scale integration of these circuit models on different SPICE solvers, including proprietary ones as long as the
SPICE circuit can be written to particular directories. However, due to the ease of circuit integration,
and translation that ``hdl21`` provides, it's API can also be used to implement parametrised functionality,
although it is easy to export the circuit as a raw SPICE directive after composition.

This composition tends to be implemented in a `SubCircuit` hierarchical implementation, as this allows for more modularisation of the netlisted devices.

Let's assume that we can get an extracted SPICE netlist of our circuit, that includes all nodes, and component
circuit definitions. This could then be simulated accordingly for the whole circuit between inputs and outputs. This
would have to be constructed out of component models and a provided netlist in a similar fashion to ``SAX``. """
# Another aspect of complexity is that `hdl21` is class-based composed, so that means that functions define class definitions and return them, or instantiate default ones.

# __all__ = ["create_pyspice_circuit", "write_raw_spice_to_pyspice_circuit"]
#
#
# def create_pyspice_circuit(circuit_name: str = "c") -> Circuit:
#     """
#     This function returns a generic PySPICE circuit. This is useful for adding SPICE directives to a circuit.
#
#     Args:
#         circuit_name (str): Name of the circuit
#
#     Returns:
#         Circuit: PySpice circuit
#     """
#     circuit = Circuit(circuit_name)
#     return circuit
#
#
# def write_raw_spice_to_pyspice_circuit(circuit: Circuit, raw_spice: str) -> Circuit:
#     """
#     This function returns a PySpice circuit including the raw SPICE text. This is useful for adding SPICE directives
#     to a circuit.
#
#     Args:
#         circuit (Circuit): PySpice circuit
#         raw_spice (str): Raw SPICE text
#
#     Returns:
#         Circuit: PySpice circuit including the raw SPICE text
#     """
#     circuit.raw_spice += raw_spice
#     return circuit



from piel.models.frequency import photonic  # NOQA: F401
from piel.models.frequency import electrical  # NOQA: F401
from piel.models.frequency import electronic  # NOQA: F401
from piel.models.frequency import electro_optic  # NOQA: F401
from piel.models.frequency import opto_electronic  # NOQA: F401
from .utils import *
from .all import get_all_models
from .defaults import get_default_models



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
import sax

__all__ = ["coupler"]


def coupler(coupling=0.5):
    kappa = coupling**0.5
    tau = (1 - coupling) ** 0.5
    sdict = sax.reciprocal(
        {
            ("in0", "out0"): tau,
            ("in0", "out1"): 1j * kappa,
            ("in1", "out0"): 1j * kappa,
            ("in1", "out1"): tau,
        }
    )
    return sdict



from typing import Literal
from .photonic.straight_waveguide import (
    waveguide,
    lossless_straight,
    ideal_active_waveguide,
)
from .photonic.mmi2x2 import mmi2x2_50_50

__all__ = [
    "get_default_models",
]

# Default model dictionary library that can be overwritten for specific modelling applications.
__default_models_dictionary__ = {
    "bend_euler": waveguide,
    "mmi2x2": mmi2x2_50_50,
    "straight": waveguide,
    "straight_heater_metal_simple": ideal_active_waveguide,
}

__default_quantum_models_dictionary__ = {
    "bend_euler": lossless_straight,
    "mmi2x2": mmi2x2_50_50,
    "straight": lossless_straight,
    "straight_heater_metal_simple": ideal_active_waveguide,
}


def get_default_models(
    custom_defaults: dict | None = None,
    type: Literal["classical", "quantum"] = "classical",
) -> dict:
    """
    Returns the default models dictionary.

    Args:
        custom_defaults (dict): Custom defaults dictionary.
        type (Literal["default", "quantum"]): Type of default models dictionary to return.

    Returns:
        dict: Default models dictionary.
    """
    if custom_defaults is not None:
        return custom_defaults
    else:
        if type == "classical":
            return __default_models_dictionary__
        elif type == "quantum":
            return __default_quantum_models_dictionary__  #
        else:
            raise ValueError(f"Type {type} not recognised.")



"""
This module aims to extend sax from standard netlist operations to include more complex operations that enable connectivity.
"""
import sax
from typing import Optional
from ...models.frequency import get_default_models

__all__ = [
    "address_value_dictionary_to_function_parameter_dictionary",
    "compose_recursive_instance_location",
    "get_component_instances",
    "get_netlist_instances_by_prefix",
    "get_matched_model_recursive_netlist_instances",
]


def address_value_dictionary_to_function_parameter_dictionary(
    address_value_dictionary: dict,
    parameter_key: str,
):
    """
    This function converts an address of an instance with particular parameter values in the form:

        {('component_lattice_gener_fb8c4da8', 'mzi_1', 'sxt'): 0,
        ('component_lattice_gener_fb8c4da8', 'mzi_5', 'sxt'): 0}

    to

        {'mzi_1': {'sxt': {parameter_key: 0}},
        ('mzi_5', {'sxt': {parameter_key: 0}}}


    """
    result = {}
    for address, value in address_value_dictionary.items():
        components = address[
            1:-1
        ]  # Exclude the first and last elements (component_lattice_gener_fb8c4da8 and sxt)
        current_dict = result
        for component in components:
            if component not in current_dict:
                current_dict[component] = {}
            current_dict = current_dict[component]
        current_dict[parameter_key] = value
    return result


def compose_recursive_instance_location(
    recursive_netlist: dict,
    top_level_instance_name: str,
    required_models: list,
    target_component_prefix: str,
    models: dict,
):
    """
       This function returns the recursive location of any matching ``target_component_prefix`` instances within the ``recursive_netlist``. A function that returns the mapping of the ``matched_component`` in the corresponding netlist at any particular level of recursion. This function iterates over a particular level of recursion of a netlist. It returns a list of the missing required components, and updates a dictionary of models that contains a particular matching component. It returns the corresponding list of instances of a particular component at that level of recursion, so that it can be appended upon in order to construct the location of the corresponding matching elements.

       If ``required_models`` is an empty list, it means no recursion is required and the function is complete. If a ``required_model_i`` in ``required_models`` matches ``target_component_prefix``, then no more recursion is required down the component function.

       The ``recursive_netlist`` should contain all the missing composed models that are not provided in the main models dictionary. If not, then we need to require the user to input the missing model that cannot be extracted from the composed netlist.
    We know when a model is composed, and when it is already provided at every level of recursion based on the ``models`` dictionary that gets updated at each level of recursion with the corresponding models of that level, and the ``required_models`` down itself.

       However, a main question appears on how to do the recursion. There needs to be a flag that determines that the recursion is complete. However, this is only valid for every particular component in the ``required_models`` list. Every component might have missing component. This means that this recursion begins component by component, updating the ``required_models`` list until all of them have been composed from the recursion or it is determined that is it missing fully.

       It would be ideal to access the particular component that needs to be implemented.

       Returns a tuple of ``model_composition_mapping, instance_composition_mapping, target_component_mapping`` in the form of

           ({'mzi_214beef3': ['straight_heater_metal_s_ad3c1693']},
            {'mzi_214beef3': ['mzi_1', 'mzi_5'],
             'mzi_d46c281f': ['mzi_2', 'mzi_3', 'mzi_4']})
    """
    model_composition_mapping = dict()
    instance_composition_mapping = dict()
    target_component_mapping = dict()
    i = 0
    while len(required_models) != 0:
        # if len(required_models) == 0:
        #     pass
        #     # Return the results as the recursive iteration is now complete.
        # else:
        # TODO Break if required_models cannot be composed and needs to be provided by the user.
        # This means that the model inside the top_level required model also has a required model that should be inside the recursive netlist and we need to find it.
        # We iterate over each of the required model names to see if they match our active component name.
        for required_model_name_i in required_models:
            # Appends required_models_i from subcomponent to the required_models input based on the models provided.
            required_models_i = sax.get_required_circuit_models(
                recursive_netlist[required_model_name_i],
                # TODO make this recursive so it can search inside? This will never have to be 2D as all models outside.
                models={**models, **model_composition_mapping},
            )  # eg. ["straight_heater_metal_s_ad3c1693"]

            # Check if required_model_name_i already composed.

            # Check that the model composition mapping has not already fulfilled this model.
            if len(required_models_i) != 0:
                if required_model_name_i in model_composition_mapping:
                    required_models.remove(required_model_name_i)
                else:
                    required_models.extend(required_models_i)
                    model_composition_mapping[required_model_name_i] = required_models_i
            elif len(required_models_i) == 0:
                # Remove from ``required_models`` to complete the recursion
                required_models.remove(required_model_name_i)

            # Get the corresponding instances of this model at this level of recursion.
            # Implement a function that matches all the potential corresponding matched instances on the top_level
            instance_composition_mapping_i = get_component_instances(
                recursive_netlist=recursive_netlist,
                top_level_prefix=top_level_instance_name,
                component_name_prefix=required_model_name_i,
            )  # {'mzi_214beef3': ['mzi_1', 'mzi_5']}
            instance_composition_mapping.update(instance_composition_mapping_i)

            # This model is now at a particular level of recursion, let's check if this is the model we want in the required composed models.
            for required_model_name_i_i in required_models_i:
                if required_model_name_i_i.startswith(target_component_prefix):
                    # Yes, this is the model we want. Can we compose the instance location?
                    target_component_mapping.update(
                        {required_model_name_i_i: required_model_name_i}
                    )
                    # This means we need to check whether the components is our matched component, and if not, then we need to check if this other required component recursively also requires our active component. Implement the search again recursively from the unmatched components.

            # If the target_component has the corresponding mapping in the recursive_netlist then we can access the lowest component element
            if required_model_name_i.startswith(target_component_prefix):
                if required_model_name_i in target_component_mapping:
                    instance_composition_mapping_i = get_component_instances(
                        recursive_netlist=recursive_netlist,
                        top_level_prefix=target_component_mapping[
                            required_model_name_i
                        ],
                        component_name_prefix=required_model_name_i,
                    )  # {'mzi_214beef3': ['mzi_1', 'mzi_5']}
                    instance_composition_mapping.update(instance_composition_mapping_i)

        i += 1

    return (
        model_composition_mapping,
        instance_composition_mapping,
        target_component_mapping,
    )


def get_component_instances(
    recursive_netlist: dict,
    top_level_prefix: str,
    component_name_prefix: str,
):
    """
    Returns a dictionary of all instances of a given component in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        top_level_prefix: The prefix of the top level instance.
        component_name_prefix: The name of the component to search for.

    Returns:
        A dictionary of all instances of the given component.
    """
    # TODO interim function until Floris merges.
    recursive_netlist_sax = sax.netlist(recursive_netlist)
    instance_names = []
    recursive_netlist_root = recursive_netlist_sax.dict()["__root__"]
    top_level_prefix = get_netlist_instances_by_prefix(
        recursive_netlist, instance_prefix=top_level_prefix
    )  # Should only be one in a netlist-to-digraph. Can always be very specified.
    for key in recursive_netlist_root[top_level_prefix]["instances"]:
        if recursive_netlist_root[top_level_prefix]["instances"][key][
            "component"
        ].startswith(component_name_prefix):
            # Note priority encoding on match.
            instance_names.append(key)
    return {component_name_prefix: instance_names}


def get_netlist_instances_by_prefix(
    recursive_netlist: dict,
    instance_prefix: str,
) -> str:
    """
    Returns a list of all instances with a given prefix in a recursive netlist.

    Args:
        recursive_netlist: The recursive netlist to search.
        instance_prefix: The prefix to search for.

    Returns:
        A list of all instances with the given prefix.
    """
    # TODO interim function until Floris merges.
    recursive_netlist = sax.netlist(recursive_netlist)
    recursive_netlist_root = recursive_netlist.dict()["__root__"]
    result = []
    for key in recursive_netlist_root.keys():
        if key.startswith(instance_prefix):
            result.append(key)

    if len(result) == 1:
        return result[0]
    elif len(result) == 0:
        raise ValueError(
            "No instances with prefix: "
            + instance_prefix
            + " found. These are the available instances: "
            + str(recursive_netlist_root.keys())
        )
    else:
        print(len(result))
        raise ValueError(
            "More than one instance with prefix: "
            + instance_prefix
            + "found. These are the matched "
            "instances: " + str(result)
        )


def get_matched_model_recursive_netlist_instances(
    recursive_netlist: dict,
    top_level_instance_prefix: str,
    target_component_prefix: str,
    models: Optional[dict] = None,
) -> list[tuple]:
    """
    This function returns an active component list with a tuple mapping of the location of the active component within the recursive netlist and corresponding model. It will recursively look within a netlist to locate what models use a particular component model. At each stage of recursion, it will compose a list of the elements that implement this matching model in order to relate the model to the instance, and hence the netlist address of the component that needs to be updated in order to functionally implement the model.

    It takes in as a set of parameters the recursive_netlist generated by a ``gdsfactory`` netlist implementation.

    Returns a list of tuples, that correspond to the phases applied with the corresponding component paths at multiple levels of recursion.
    eg. [("component_lattice_gener_fb8c4da8", "mzi_1", "sxt"), ("component_lattice_gener_fb8c4da8", "mzi_5", "sxt")] and these are our keys to our sax circuit decomposition.
    """
    matched_instance_list = []
    if models is None:
        models = get_default_models()

    # We need to input the top-level instance.
    top_level_instance_name = get_netlist_instances_by_prefix(
        recursive_netlist=recursive_netlist,
        instance_prefix=top_level_instance_prefix,
    )

    # We need to input the prefix of the component of the straight metal heater.
    top_level_required_models = sax.get_required_circuit_models(
        recursive_netlist[top_level_instance_name],
        models=models,
    )

    (
        model_composition_mapping,
        instance_composition_mapping,
        target_component_mapping,
    ) = compose_recursive_instance_location(
        recursive_netlist=recursive_netlist,
        top_level_instance_name=top_level_instance_name,
        required_models=top_level_required_models.copy(),
        target_component_prefix=target_component_prefix,
        models=models,
    )

    # Now we have the raw data that creates the mapping of the components-to-instances, in order to create the corresponding instance address indexes that we can use to control our matching element parameters.
    if len(target_component_mapping.keys()) != 0:
        # This means that the target_component has been mapped to a parent recursive_netlist cell.
        for target_component_name_i in target_component_mapping.keys():
            # Tells us the name of our component
            recursive_parent_component_i = target_component_mapping[
                target_component_name_i
            ]  # Get the parent cell.
            for parent_instances_i in instance_composition_mapping[
                recursive_parent_component_i
            ]:
                # TODO check parent_instances_i not in target_component_mapping to increase hierarchy.
                # TODO implement as another recursive problem. NMP right now.
                for target_instances_i in instance_composition_mapping[
                    target_component_name_i
                ]:
                    matched_instance_list.append(
                        (
                            top_level_instance_name,
                            parent_instances_i,
                            target_instances_i,
                        )
                    )
    return matched_instance_list



__all__ = ["linear_phase_mapping_relationship"]


def linear_phase_mapping_relationship(
    phase_power_slope: float,
    zero_power_phase: float,
):
    """
    This function returns a function that maps the power applied to a particular heater resistor linearly. For
    example, we might start with a minimum phase mapping of (0,0) where the units are in (Watts, Phase). If we have a ridiculous arbitrary phase_power_slope of 1rad/1W, then the points in our linear mapping would be (0,0), (1,1), (2,2), (3,3), etc. This is implemented as a lambda function that takes in a power and returns a phase. The units of the power and phase are determined by the phase_power_slope and zero_power_phase. The zero_power_phase is the phase at zero power. The phase_power_slope is the slope of the linear mapping. The units of the phase_power_slope are radians/Watt. The units of the zero_power_phase are radians. The units of the power are Watts. The units of the phase are radians.

    Args:
        phase_power_slope (float): The slope of the linear mapping. The units of the phase_power_slope are radians/Watt.
        zero_power_phase (float): The phase at zero power. The units of the zero_power_phase are radians.

    Returns:
        linear_phase_mapping (function): A function that maps the power applied to a particular heater resistor linearly. The units of the power and phase are determined by the phase_power_slope and zero_power_phase. The zero_power_phase is the phase at zero power. The phase_power_slope is the slope of the linear mapping. The units of the phase_power_slope are radians/Watt. The units of the zero_power_phase are radians. The units of the power are Watts. The units of the phase are radians.
    """

    def linear_phase_mapping(power_w: float) -> float:
        """
        We create a linear interpolation based on the phase_power_slope. This function returns phase in radians.
        """
        return phase_power_slope * power_w + zero_power_phase

    return linear_phase_mapping



"""
These are the corresponding prefixes from `hdl21`:

f = FEMTO = Prefix.FEMTO
p = PICO = Prefix.PICO
n = NANO = Prefix.NANO
µ = u = MICRO = Prefix.MICRO # Note both `u` and `µ` are valid
m = MILLI = Prefix.MILLI
K = KILO = Prefix.KILO
M = MEGA = Prefix.MEGA
G = GIGA = Prefix.GIGA
T = TERA = Prefix.TERA
P = PETA = Prefix.PETA
UNIT = Prefix.UNIT

"""
import hdl21 as h
import numpy as np

__all__ = ["convert_numeric_to_prefix"]


def convert_numeric_to_prefix(
    value: float,
):
    """
    This function converts a numeric value to a number under a SPICE unit closest to the base prefix. This allows us to connect a particular number real output, into a term that can be used in a SPICE netlist.
    """
    prefixes = [
        (h.Prefix.YOCTO, h.prefix.y),
        (h.Prefix.ZEPTO, h.prefix.z),
        (h.Prefix.ATTO, h.prefix.a),
        (h.Prefix.FEMTO, h.prefix.f),
        (h.Prefix.PICO, h.prefix.p),
        (h.Prefix.NANO, h.prefix.n),
        (h.Prefix.MICRO, h.prefix.µ),
        (h.Prefix.MICRO, h.prefix.u),
        (h.Prefix.MILLI, h.prefix.m),
        (h.Prefix.CENTI, h.prefix.c),
        (h.Prefix.DECI, h.prefix.d),
        # (Prefix.UNIT, ''),
        (h.Prefix.DECA, h.prefix.D),
        (h.Prefix.KILO, h.prefix.K),
        (h.Prefix.MEGA, h.prefix.M),
        (h.Prefix.GIGA, h.prefix.G),
        (h.Prefix.TERA, h.prefix.T),
        (h.Prefix.PETA, h.prefix.P),
        (h.Prefix.EXA, h.prefix.E),
        (h.Prefix.ZETTA, h.prefix.Z),
        (h.Prefix.YOTTA, h.prefix.Y),
    ]

    base_10 = np.log10(value)
    value_target_base = np.floor(base_10)

    closest_prefix = None
    min_difference = 2
    for prefix, _ in prefixes:
        difference = abs(value_target_base - prefix.value)
        if difference < min_difference:
            min_difference = difference
            closest_prefix = prefix

    value /= 10**closest_prefix.value

    return value * closest_prefix



"""
Hi Dan,

Hope you are well.
"""
import hdl21 as h


@h.paramclass
class MyParams:
    # Required
    width = h.Param(dtype=int, desc="Width. Required", default=10)
    # Optional - including a default value
    text = h.Param(dtype=str, desc="Optional string", default="My Favorite Module")


@h.generator
def MyFirstGenerator(params: MyParams) -> h.Module:
    # A very exciting first generator function
    print("MyFirstGenerator")
    m = h.Module()
    m.i = h.Input(width=params.width)
    return m


@h.generator
def MySecondGenerator(params: MyParams) -> h.Module:
    # A very exciting first generator function
    m = h.Module()
    m.i = h.Input(width=params.width)
    m.a = MyFirstGenerator(width=params.width)(**{"i": m.i})
    # print(getattr(m, "i"))
    print("MyFirstGenerator ports")
    # m.i = m.a.i
    return m


a = MySecondGenerator(width=10)
print(a)
b = h.elaborate(a)
print(b)
# import sys

# from hdl21.elab.elaborators.conntypes import get_unconnected_instance_connections

# missing_conns = get_unconnected_instance_connections(b, inst=getattr(b, "a"))
# print(missing_conns)
# print(h.netlist(b, sys.stdout, fmt="spice"))


"""
RuntimeError: Elaboration Error at hierarchical path:
  Module        straight_heater_metal_s_b8a2a400s
  Instance      via_stack_1
Invalid connections `{'e1': Missing connection to Port `e1`, 'e2': Missing connection to Port `e2`, 'e4': Missing connection to Port `e4`}` on Instance `via_stack_1` in Module `straight_heater_metal_s_b8a2a400`

According to HDL21 currently, the existing internal ports do not solve external connectivity requirements in order to generate the netlist. This means that the declared ports at the top level must be connected. This means that we need a subroutine to fix this, and ideally it would be done by the generators but this is part of the composition. It would be good to have a helper function to extract this in terms of the composition. This may not be able to be done in the gdsfactory netlist, due to the nature of the connectivity. What would make sense is to expose them.

Part of the complexity is that the GeneratorCall class does not expose its ports until it is generated. So it can connect hypothetical ports, but it cannot expose the missing ports until after generation. We can use the unconencted ports warning though.
"""



"""
These functions aim to provide functionality to parse data from any OpenLanes v1 design into Python.

They are ported from the old: github.com/daquintero/porf
"""
from .run_output import *
from .sta_rpt import *
from .utils import *



from typing import Literal, Optional

__all__ = ["get_input_ports_index", "get_matched_ports_tuple_index"]


def get_matched_ports_tuple_index(
    ports_index: dict,
    selected_ports_tuple: Optional[tuple] = None,
    sorting_algorithm: Literal["prefix", "selected_ports"] = "prefix",
    prefix: str = "in",
) -> (tuple, tuple):
    """
    This function returns the input ports of a component. However, input ports may have different sets of prefixes
    and suffixes. This function implements different sorting algorithms for different ports names. The default
    algorithm is `prefix`, which sorts the ports by their prefix. The Endianness implementation means that the tuple
    order is determined according to the last numerical index order of the port numbering. Returns just a tuple of
    the index.

    .. code-block:: python

        raw_ports_index = {
            "in_o_0": 0,
            "out_o_0": 1,
            "out_o_1": 2,
            "out_o_2": 3,
            "out_o_3": 4,
            "in_o_1": 5,
            "in_o_2": 6,
            "in_o_3": 7,
        }

        get_input_ports_tuple_index(ports_index=raw_ports_index)

        # Output
        (0, 5, 6, 7)

    Args:
        ports_index (dict): The ports index dictionary.
        selected_ports_tuple (tuple, optional): The selected ports tuple. Defaults to None.
        sorting_algorithm (Literal["prefix"], optional): The sorting algorithm to use. Defaults to "prefix".
        prefix (str, optional): The prefix to use for the sorting algorithm. Defaults to "in".

    Returns:
        matches_ports_index_tuple_order(tuple): The ordered input ports index tuple.
        matched_ports_name_tuple_order(tuple): The ordered input ports name tuple.

    """
    # TODO optimise this computaiton
    ports_index_list = list(ports_index.keys())
    matches_ports_index_tuple_order = tuple()
    matched_ports_list = tuple()
    if sorting_algorithm == "prefix":
        matched_ports_list = [
            item for item in ports_index_list if item.startswith(prefix)
        ]
        # This sorts in numerical order
        matched_ports_list.sort()
        matched_ports_name_tuple_order = tuple(matched_ports_list)
        matches_ports_index_tuple_order = tuple(
            [ports_index[port] for port in matched_ports_list]
        )
    elif sorting_algorithm == "selected_ports":
        matched_ports_list = selected_ports_tuple
        matched_ports_name_tuple_order = tuple(matched_ports_list)
        matches_ports_index_tuple_order = tuple(
            [ports_index[port] for port in matched_ports_list]
        )
    else:
        raise ValueError(
            "matches_ports_index_tuple_order "
            + str(matches_ports_index_tuple_order)
            + " not defined."
        )
    return matches_ports_index_tuple_order, matched_ports_name_tuple_order


def get_input_ports_index(
    ports_index: dict,
    sorting_algorithm: Literal["prefix"] = "prefix",
    prefix: str = "in",
) -> tuple:
    """
    This function returns the input ports of a component. However, input ports may have different sets of prefixes and suffixes. This function implements different sorting algorithms for different ports names. The default algorithm is `prefix`, which sorts the ports by their prefix. The Endianness implementation means that the tuple order is determined according to the last numerical index order of the port numbering.

    .. code-block:: python

        raw_ports_index = {
            "in_o_0": 0,
            "out_o_0": 1,
            "out_o_1": 2,
            "out_o_2": 3,
            "out_o_3": 4,
            "in_o_1": 5,
            "in_o_2": 6,
            "in_o_3": 7,
        }

        get_input_ports_index(ports_index=raw_ports_index)

        # Output
        ((0, "in_o_0"), (5, "in_o_1"), (6, "in_o_2"), (7, "in_o_3"))

    Args:
        ports_index (dict): The ports index dictionary.
        sorting_algorithm (Literal["prefix"], optional): The sorting algorithm to use. Defaults to "prefix".
        prefix (str, optional): The prefix to use for the sorting algorithm. Defaults to "in".

    Returns:
        tuple: The ordered input ports index tuple.
    """
    # TODO optimise this computaiton
    ports_index_order = tuple()
    if sorting_algorithm == "prefix":
        matched_ports_index_tuple, matched_ports_list = get_matched_ports_tuple_index(
            ports_index=ports_index, sorting_algorithm=sorting_algorithm, prefix=prefix
        )
        ports_index_order = tuple(
            zip(matched_ports_index_tuple, matched_ports_list, strict=True)
        )

    return ports_index_order



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
import jax.numpy as jnp

__all__ = ["grating_coupler_simple"]


def grating_coupler_simple(R=0.0, R_in=0.0, Tmax=1.0, bandwidth=0.06e-6, wl0=1.55e-6):
    # Constants
    fwhm2sigma = 1.0 / (2 * jnp.sqrt(2 * jnp.log(2)))

    # Compute sigma
    sigma = fwhm2sigma * bandwidth

    # Assume the wavelength of the environment matches the center wavelength of the grating coupler
    wls = wl0

    # Compute loss
    loss = jnp.sqrt(Tmax * jnp.exp(-((wl0 - wls) ** 2) / (2 * sigma**2)))

    # Create scattering dictionary
    sdict = {
        ("in0", "out1"): loss,
        ("in1", "out0"): loss,
        ("in0", "out0"): R_in,
        ("in1", "out1"): R,
    }

    return sdict



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
# def realistic_directional_coupler(length=12.8e-6, k0=0.2332, n0=0.0208, de1_k0=1.2435, de1_n0=0.1169, de2_k0=5.3022, de2_n0=0.4821, wl0=1.55e-6):
#     wl = torch.tensor(1.55e-6, dtype=torch.float64)
#     dwl = wl - wl0
#     dn = n0 + de1_n0 * dwl + 0.5 * de2_n0 * dwl**2
#     kappa0 = k0 + de1_k0 * dwl + 0.5 * de2_k0 * dwl**2
#     kappa1 = np.pi * dn / wl
#     tau = torch.cos(kappa0 + kappa1 * length)
#     kappa = -torch.sin(kappa0 + kappa1 * length)
#     sdict = sax.reciprocal({
#         ("port0", "port1"): tau,
#         ("port0", "port2"): 1j*kappa,
#         ("port1", "port3"): 1j*kappa,
#         ("port2", "port3"): tau,
#     })
#     return sdict



import pandas as pd
import pathlib
from piel.file_system import check_path_exists, return_path

__all__ = [
    "contains_in_lines",
    "create_file_lines_dataframe",
    "get_file_line_by_keyword",
    "read_file_lines",
]


def contains_in_lines(
    file_lines_data: pd.DataFrame,
    keyword: str,
):
    """
    Check if the keyword is contained in the file lines

    Args:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines
        keyword (str): Keyword to search for

    Returns:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines
    """
    return file_lines_data.lines.str.contains(keyword)


def create_file_lines_dataframe(file_lines_raw):
    """
    Create a DataFrame from the raw lines of a file

    Args:
        file_lines_raw (list): list containing the file lines

    Returns:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines
    """
    return pd.DataFrame({"lines": file_lines_raw})


def get_file_line_by_keyword(
    file_lines_data: pd.DataFrame,
    keyword: str,
    regex: str,
):
    """
    Extract the data from the file lines using the given keyword and regex

    Args:
        file_lines_data (pd.DataFrame): Dataframe containing the file lines
        keyword (str): Keyword to search for
        regex (str): Regex to extract the data

    Returns:
        extracted_values (pd.DataFrame): Dataframe containing the extracted values
    """
    lines_with_keyword = file_lines_data.lines[file_lines_data[f"{keyword}_line"]]
    extracted_values = lines_with_keyword.str.extract(regex)
    return extracted_values


def read_file_lines(file_path: str | pathlib.Path):
    """
    Extract lines from the file

    Args:
        file_path (str | pathlib.Path): Path to the file

    Returns:
        file_lines_raw (list): list containing the file lines
    """
    file_path = return_path(file_path)
    file = read_file(file_path)
    return file.readlines()


def read_file(file_path: str | pathlib.Path):
    """
    Read the file from the given path

    Args:
        file_path (str | pathlib.Path): Path to the file

    Returns:
        file: the opened file
    """
    file_path = return_path(file_path)
    check_path_exists(file_path)
    return open(str(file_path.resolve()), "r")



from .taper import taper
from .straight import straight
from .via_stack import via_stack

__all__ = ["get_default_models"]


__default_models_dictionary__ = {
    "taper": taper,
    "straight": straight,
    "via_stack": via_stack,
}


def get_default_models(custom_defaults: dict | None = None) -> dict:
    """
    Returns the default models dictionary.

    Args:
        custom_defaults (dict): Custom defaults dictionary.

    Returns:
        dict: Default models dictionary.
    """
    if custom_defaults is not None:
        return custom_defaults
    else:
        return __default_models_dictionary__



"""
This file contains a range of functions used to read, plot and analyse cocotb simulations in a data-flow standard as suggested
"""
import functools
import pandas as pd
from piel.config import piel_path_types
from piel.file_system import return_path, get_files_recursively_in_directory

__all__ = [
    "get_simulation_output_files",
    "get_simulation_output_files_from_design",
    "read_simulation_data",
    "simple_plot_simulation_data",
]

get_simulation_output_files = functools.partial(
    get_files_recursively_in_directory, path="./tb/out/", extension="csv"
)


def get_simulation_output_files_from_design(
    design_directory: piel_path_types,
    extension: str = "csv",
):
    """
    This function returns a list of all the simulation output files in the design directory.

    Args:
        design_directory (piel_path_types): The path to the design directory.

    Returns:
        output_files (list): List of all the simulation output files in the design directory.
    """
    design_directory = return_path(design_directory)
    output_files = get_files_recursively_in_directory(
        path=design_directory / "tb" / "out", extension=extension
    )
    return output_files


def read_simulation_data(file_path: piel_path_types):
    """
    This function returns a Pandas dataframe that contains all the simulation data outputted from the simulation run.

    Args:
        file_path (piel_path_types): The path to the simulation data file.

    Returns:
        simulation_data (pd.DataFrame): The simulation data in a Pandas dataframe.
    """
    file_path = return_path(file_path)
    simulation_data = pd.read_csv(file_path)
    return simulation_data


def simple_plot_simulation_data(simulation_data: pd.DataFrame):
    from bokeh.models import ColumnDataSource
    from bokeh.plotting import figure, show
    from bokeh.layouts import column

    source = ColumnDataSource(
        data=dict(time=simulation_data.t, signal=simulation_data.x)
    )

    p = figure(
        height=300,
        width=800,
        tools="xpan",
        toolbar_location=None,
        # x_axis_type="datetime",
        x_axis_location="above",
        background_fill_color="#efefef",
    )

    p.line("time", "signal", source=source)
    # p.yaxis.axis_label = "Price"

    select = figure(
        title="Drag the middle and edges of the selection box to change the range above",
        height=130,
        width=800,
        y_range=p.y_range,
        # x_axis_type="datetime",
        y_axis_type=None,
        tools="",
        toolbar_location=None,
        background_fill_color="#efefef",
    )

    # range_tool = RangeTool(x_range=p.x_range)
    # range_tool.overlay.fill_color = "navy"
    # range_tool.overlay.fill_alpha = 0.2

    select.line("time", "signal", source=source)
    select.ygrid.grid_line_color = None
    # select.add_tools(range_tool)

    return show(column(p, select))



from .coupler_simple import *
from .directional_coupler_length import *
from .directional_coupler_real import *
from .directional_coupler_simple import *
from .grating_coupler import *
from .mmi1x2 import *
from .mmi2x2 import *
from .straight_waveguide import *



import hdl21 as h
import hdl21.sim as hs
import numpy as np
import pandas as pd
from typing import Literal, Optional
import vlsirtools.spice as vsp
from ...config import piel_path_types
from ...file_system import return_path

__all__ = [
    "configure_ngspice_simulation",
    "configure_operating_point_simulation",
    "configure_transient_simulation",
    "run_simulation",
]


def configure_ngspice_simulation(
    run_directory: piel_path_types = ".",
):
    """
    This function configures the NGSPICE simulation for the circuit and returns a simulation class.

    Args:
        run_directory (piel_path_types): Directory where the simulation will be run

    Returns:
        simulation_options: Configured NGSPICE simulation options
    """
    run_directory = return_path(run_directory)
    simulation_options = vsp.SimOptions(
        simulator=vsp.SupportedSimulators.NGSPICE,
        fmt=vsp.ResultFormat.SIM_DATA,
        rundir=run_directory,
    )
    return simulation_options


def configure_operating_point_simulation(
    testbench: h.Module,
    **kwargs,
):
    """
    This function configures the DC operating point simulation for the circuit and returns a simulation class.

    Args:
        testbench (Module): HDL21 testbench
        **kwargs: Additional arguments to be passed to the operating point simulation such as name.

    Returns:
        Simulation: HDL21 simulation class
    """

    @hs.sim
    class Simulation:
        tb = testbench
        operating_point_tb = hs.Op(**kwargs)

    return Simulation


def configure_transient_simulation(
    testbench: h.Module,
    stop_time_s: float,
    step_time_s: float,
    **kwargs,
):
    """
    This function configures the transient simulation for the circuit and returns a simulation class.

    Args:
        testbench (Module): HDL21 testbench
        stop_time_s (float): Stop time of the simulation in seconds
        step_time_s (float): Step time of the simulation in seconds
        **kwargs: Additional arguments to be passed to the transient simulation

    Returns:
        Simulation: HDL21 simulation class
    """

    @hs.sim
    class Simulation:
        tb = testbench
        transient_tb = hs.Tran(
            tstop=stop_time_s * h.prefix.UNIT,
            tstep=step_time_s * h.prefix.UNIT,
            **kwargs,
        )

    return Simulation


def save_results_to_csv(
    results: hs.SimResult, file_name: str, save_directory: piel_path_types = "."
):
    """
    This function converts the simulation results to a pandas dataframe and saves it to a csv file.

    Args:
        directory (piel_path_types): Directory where the simulation will be run
    """

    save_directory = return_path(save_directory)
    # TODO check that there are more than one analysis
    analysis_results = results.an[0].data
    if type(next(iter(analysis_results.values()))) not in (
        list,
        dict,
        tuple,
        np.ndarray,
    ):
        # Check that dict values are scalars
        analysis_results = pd.DataFrame(analysis_results, index=[0])
    else:
        analysis_results = pd.DataFrame(analysis_results)

    analysis_results.to_csv(save_directory / (file_name + ".csv"))


def run_simulation(
    simulation: h.sim.Sim,
    simulator_name: Literal["ngspice"] = "ngspice",
    simulation_options: Optional[vsp.SimOptions] = None,
    to_csv: bool = True,
):
    """
    This function runs the transient simulation for the circuit and returns the results.

    Args:
        simulation (h.sim.Sim): HDL21 simulation class
        simulator_name (Literal["ngspice"]): Name of the simulator
        simulation_options (Optional[vsp.SimOptions]): Simulation options
        to_csv (bool): Whether to save the results to a csv file

    Returns:
        results: Simulation results
    """
    if simulator_name == "ngspice":
        if simulation_options is None:
            if not vsp.ngspice.available():
                print(
                    "NGSPICE is not available. Please install it. Check by running `ngspice` in the terminal."
                )
                return
            simulation_options = configure_ngspice_simulation()
        results = simulation.run(simulation_options)
    else:
        print("Simulator not supported.")
        return

    if to_csv:
        save_results_to_csv(results, simulation.tb.name)

    return results



"""
SPICE Resistor Structure

See Mike Smith “WinSpice3 User’s Manual” 25 October, 1999

.. code-block:: spice

    RXXXXXXX N1 N2 <VALUE> <MNAME> <L=LENGTH> <W=WIDTH> <TEMP=T>

Where the terminals are:

.. code-block::

    N1 = the first terminal
    N2 = the second terminal
    <VALUE> = resistance in ohms.
    <MNAME> = name of the model used (useful for semiconductor resistors)
    <L=LENGTH> = length of the resistor (useful for semiconductor resistors)
    <W=WIDTH> = width of the resistor (useful for semiconductor resistors)
    <TEMP=T> = temperature of the resistor in Kelvin (useful in noise analysis and
    semiconductor resistors)

An example is:

.. code-block::

    RHOT n1 n2 10k TEMP=500

"""



"""
# TODO move and update.
The way each of these models should work is that they use the settings from the `gdsfactory` component,
to create a parametric SPICE directive.

These functions map a particular model, with an instance representation that corresponds to the given netlist
connectivity, and returns a SPICE representation of the circuit. This function will be called after parsing the
circuit netlist accordingly, and creating a mapping from the instance definitions to the fundamental components.
"""


def add_basic_capacitor(settings) -> str:
    """
    This function takes in the settings from a gdsfactory component, some connectivity node translated directly from
    the gdsfactory netlist.

    See Mike Smith “WinSpice3 User’s Manual” 25 October, 1999

    SPICE capacitor model:

    .. code-block::

        CXXXXXXX N+ N- VALUE <IC=INCOND>

    Where the parameters are:

    .. code-block::

        N+ = the positive terminal
        N- = the negative terminal
        VALUE = capacitance in farads
        <IC=INCOND> = starting voltage in a simulation

    """
    pass



"""
We create a set of parameters that can be used throughout the project for optimisation.

The numerical solver is jax and is imported throughout the module.
"""
import os
import pathlib
import types

__all__ = [
    "piel_path_types",
]

piel_path_types = str | pathlib.Path | os.PathLike | types.ModuleType



"""Most of the ``hdl21``-``gdsfactory`` integration functions will be contributed directly to `gdsfactory`. However,
some `translation language` inherent to the ``piel`` implementation of these tools is included here.

Note that to be able to construct a full circuit model of the netlist tools provided, it is necessary to create
individual circuit models of the devices that we will interconnect, and then map them to a larger netlist. This means
that it is necessary to create specific SPICE models for each particular component, say in an electrical netlist.

This functions convert a GDSFactory netlist, with a set of component models, into `hdl21` that accounts for the
instance properties, which can then be connected into a VLSIR compatible `Netlist` implementation.

Eventually we will implement RCX where we can extract the netlist with parasitics directly from the layout,
but for now this will be the implementation. The output structure of our SPICE should be compatible with the
`Netlist` package BaseModel.

We follow the principle in: https://eee.guc.edu.eg/Courses/Electronics/ELCT503%20Semiconductors/Lab/spicehowto.pdf

.. code-block:: spice

    Spice Simulation 1-1
    *** MODEL Descriptions ***
    .model nm NMOS level=2 VT0=0.7
    KP=80e-6 LAMBDA=0.01

    *** NETLIST Description ***
    M1 vdd ng 0 0 nm W=3u L=3u
    R1 in ng 50
    Vdd vdd 0 5
    Vin in 0 2.5

    *** SIMULATION Commands ***
    .op
    .end

Note that the netlist device connectivity structure of most passive components is in the form:

.. code-block:: spice

    <DEVICE ID> <CONNECTION_0> <CONNECTION_1> <DEVICE_VALUE> <MORE_PARAMETERS>

Our example GDSFactory netlist format is in the simplified form:

.. code-block::

    {
        "connections": {
            "straight_1": {
                "e1": "taper_1,e2",
                "e2": "taper_2,e2"
            },
            "taper_1": {
                "e1": "via_stack_1,e3"
            },
            "taper_2": {
                "e1": "via_stack_2,e1"
            }
        },
        "instances": {
            "straight_1": {
                "component": "straight",
                "info": {
                    "length": 15.0,
                    "width": 0.5,
                    "cross_section": "strip_heater_metal",
                    "settings": {
                        "width": 0.5,
                        "layer": "WG",
                        "heater_width": 2.5,
                        "layer_heater": "HEATER"
                    }
                }
            },
            "taper_1": {
                "component": "taper",
                "info": {
                    "length": 5.0,
                    "width1": 11.0,
                    "width2": 2.5
                },
                "settings": {
                    "cross_section": {
                        "layer": "HEATER",
                        "width": 2.5,
                        "offset": 0.0,
                        "taper_length": 10.0,
                        "gap": 5.0,
                        "min_length": 5.0,
                        "port_names": ["e1", "e2"]
                    }
                }
            },
            "taper_2": {
                "component": "taper",
                "info": {
                    "length": 5.0,
                    "width1": 11.0,
                    "width2": 2.5
                },
                "settings": {
                    "cross_section": {
                        "layer": "HEATER",
                        "width": 2.5,
                        "offset": 0.0,
                        "taper_length": 10.0,
                        "gap": 5.0,
                        "min_length": 5.0,
                        "port_names": ["e1", "e2"]
                    }
                }
            },
            "via_stack_1": {
                "component": "via_stack",
                "info": {
                    "size": [11.0, 11.0],
                    "layer": "M3"
                },
                "settings": {
                    "layers": ["HEATER", "M2", "M3"]
                }
            },
            "via_stack_2": {
                "component": "via_stack",
                "info": {
                    "size": [11.0, 11.0],
                    "layer": "M3"
                },
                "settings": {
                    "layers": ["HEATER", "M2", "M3"]
                }
            }
        },
        "placements": {
            "straight_1": {"x": 0.0, "y": 0.0, "rotation": 0, "mirror": 0},
            "taper_1": {"x": -5.0, "y": 0.0, "rotation": 0, "mirror": 0},
            "taper_2": {"x": 20.0, "y": 0.0, "rotation": 180, "mirror": 0},
            "via_stack_1": {"x": -10.5, "y": 0.0, "rotation": 0, "mirror": 0},
            "via_stack_2": {"x": 25.5, "y": 0.0, "rotation": 0, "mirror": 0}
        },
        "ports": {
            "e1": "taper_1,e2",
            "e2": "taper_2,e2"
        },
        "name": "straight_heater_metal_simple",
    }

This is particularly useful when creating our components and connectivity, because what we can do is instantiate our
devices with their corresponding values, and then create our connectivity accordingly. To do this properly from our
GDSFactory netlist to ``hdl21``, we can then extract the total SPICE circuit, and convert it to a VLSIR format using
the ``Netlist`` module. The reason why we can't use the Netlist package from Dan Fritchman directly is that we need to
apply a set of models that translate a particular component instantiation into an electrical model. Because we are
not yet doing layout extraction as that requires EM solvers, we need to create some sort of SPICE level assignment
based on the provided dictionary.

Note that ``hdl21`` already can implement the port connectivity directly from internal instances, and translate this
to our connectivity netlist. This means we only need to iterate to create our instances based on our models into a
``hdl21`` module, then we can easily assign the corresponding values. It is also possible to create the assigned
parameters as part of the ``hdl21`` component which would form part of our module. Because the gdsfactory names are
compatible with ``hdl21``, then it is fine to create the integration accordingly.

The algorithm can be to:

1. Parse the gdsfactory netlist, assign the electrical ports for the model. Extract all instances and
required models from the netlist.
2. Verify that the models have been provided. Each model describes the type of
component this is, how many ports it requires and so on. Create a ``hdl21`` top level module for every gdsfactory
netlist, this is reasonable as it is composed, and not a generator class. This generates a large amount of instantiated ``hdl21`` modules that are generated from `generators`.
3. Map the connections to each instance port as part of the instance dictionary. This parses the connectivity in the ``gdsfactory`` netlist and connects the ports accordingly.

The connections are a bit more complex. So each of our connections dictionary is in the form:

.. code-block::

     "connections": {
                "straight_1": {
                    "e1": "taper_1,e2",
                    "e2": "taper_2,e2"
                },
                "taper_1": {
                    "e1": "via_stack_1,e3"
                },
                "taper_2": {
                    "e1": "via_stack_2,e1"
                }
            },

We know what our top model ports are. We know our internal instance ports as well, and this will be provided by the
model too. For the sake of easiness, we can describe these as ``hdl21`` equivalent ``InOut`` or ``Port` `ports and
not have to deal with directionality. After instance declaration, and models for each of these components with the
corresponding port topology, it is then straightforward to parse the connectivity and implement the network,
and extract the SPICE. """
import hdl21
import hdl21 as h
from .conversion import (
    convert_connections_to_tuples,
    gdsfactory_netlist_with_hdl21_generators,
)

__all__ = ["gdsfactory_netlist_to_spice_netlist", "construct_hdl21_module"]


def gdsfactory_netlist_to_spice_netlist(
    gdsfactory_netlist: dict, generators: dict, **kwargs
) -> hdl21.Module:
    """
    This function converts a GDSFactory electrical netlist into a standard SPICE netlist. It follows the same
    principle as the `sax` circuit composition.

    Each GDSFactory netlist has a set of instances, each with a corresponding model, and each instance with a given
    set of geometrical settings that can be applied to each particular model. We know the type of SPICE model from
    the instance model we provides.

    We know that the gdsfactory has a set of instances, and we can map unique models via sax through our own
    composition circuit. Write the SPICE component based on the model into a total circuit representation in string
    from the reshaped gdsfactory dictionary into our own structure.

    Args:
        gdsfactory_netlist: GDSFactory netlist
        generators: Dictionary of Generators

    Returns:
        hdl21 module or raw SPICE string
    """
    spice_netlist = gdsfactory_netlist_with_hdl21_generators(
        gdsfactory_netlist=gdsfactory_netlist, generators=generators
    )
    hdl21_module = construct_hdl21_module(spice_netlist=spice_netlist)
    return hdl21_module


def construct_hdl21_module(spice_netlist: dict, **kwargs) -> h.Module:
    """
    This function converts a gdsfactory-spice converted netlist using the component models into a SPICE circuit.

    Part of the complexity of this function is the multiport nature of some components and models, and assigning the
    parameters accordingly into the SPICE function. This is because not every SPICE component will be bi-port,
    and many will have multi-ports and parameters accordingly. Each model can implement the composition into a
    SPICE circuit, but they depend on a set of parameters that must be set from the instance. Another aspect is
    that we may want to assign the component ID according to the type of component. However, we can also assign the
    ID based on the individual instance in the circuit, which is also a reasonable approximation. However,
    it could be said, that the ideal implementation would be for each component model provided to return the SPICE
    instance including connectivity except for the ID.

    # TODO implement validators
    """
    circuit = h.Module(name=spice_netlist["name"])
    instance_id = 0
    # Declare all the instances
    for instance_name_i, instance_settings_i in spice_netlist["instances"].items():
        instance_i = instance_settings_i["hdl21_model"](
            name=instance_name_i, **instance_settings_i["settings"]
        )()
        circuit.add(val=instance_i, name=instance_name_i)
        instance_id += 1

    # Create top level ports
    for port_name_i, _ in spice_netlist["ports"].items():
        # TODO include directionality on port_settings so that it can be easily interconencted with hdl21
        circuit.add(val=h.Port(name=port_name_i))

    # Create the connectivity
    connections_list = convert_connections_to_tuples(spice_netlist["connections"])
    for connection_tuple in connections_list:
        # Connects the corresponding ports.
        first_instance = getattr(circuit, connection_tuple[0][0])
        second_instance = getattr(circuit, connection_tuple[1][0])
        first_port_name = connection_tuple[0][1]
        second_port = getattr(second_instance, connection_tuple[1][1])
        first_instance.connect(first_port_name, second_port)

    # Expose all the missing electrical warning connection internal ports to the outer circuit composition so that full modelling can be performed, and no construction errors.
    if "warnings" in spice_netlist:
        if "electrical" in spice_netlist["warnings"]:
            if "unconnected_ports" in spice_netlist["warnings"]["electrical"]:
                unconnected_ports_list = spice_netlist["warnings"]["electrical"][
                    "unconnected_ports"
                ][0]["ports"]
                for instance_port_name_i in unconnected_ports_list:
                    instance_port_spice_name_i = instance_port_name_i.replace(",", "__")
                    instance_name_i, port_name_i = instance_port_name_i.split(",")
                    instance_i = getattr(circuit, instance_name_i)
                    circuit.add(val=h.Port(), name=instance_port_spice_name_i)
                    circuit_port_i = getattr(circuit, instance_port_spice_name_i)
                    instance_i.connect(port_name_i, circuit_port_i)

    # Create the top level connectivity between the top circuit ports to the instances ports.
    for circuit_port_name_i, instance_port_name_raw_i in spice_netlist["ports"].items():
        instance_name_i, instance_port_name_i = instance_port_name_raw_i.split(",")
        instance_i = getattr(circuit, instance_name_i)
        circuit_port_i = getattr(circuit, circuit_port_name_i)
        instance_i.connect(instance_port_name_i, circuit_port_i)

    return h.elaborate(circuit)



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
import sax

__all__ = ["directional_coupler"]


def directional_coupler(coupling=0.5):
    kappa = coupling**0.5
    tau = (1 - coupling) ** 0.5
    sdict = sax.reciprocal(
        {
            ("port0", "port1"): tau,
            ("port0", "port2"): 1j * kappa,
            ("port1", "port3"): 1j * kappa,
            ("port2", "port3"): tau,
        }
    )
    return sdict



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
import jax.numpy as jnp
import sax

__all__ = ["directional_coupler_with_length"]


def directional_coupler_with_length(
    length=1e-5, coupling=0.5, loss=0, neff=2.34, wl0=1.55e-6, ng=3.40, phase=0
):
    kappa = coupling**0.5
    tau = (1 - coupling) ** 0.5
    loss = 10 ** (-loss * length / 20)  # factor 20 bc amplitudes, not intensities.
    cos_phase = jnp.cos(phase)
    sin_phase = jnp.sin(phase)
    sdict = sax.reciprocal(
        {
            ("port0", "port1"): tau * loss * cos_phase,
            ("port0", "port2"): -kappa * loss * sin_phase,
            ("port1", "port3"): -kappa * loss * sin_phase,
            ("port2", "port3"): tau * loss * cos_phase,
        }
    )
    return sdict



import jax.numpy as jnp
import numpy as np
from typing import Optional
import qutip

__all__ = [
    "standard_s_parameters_to_qutip_qobj",
    "verify_matrix_is_unitary",
    "subunitary_selection_on_range",
    "subunitary_selection_on_index",
]


def matrix_to_qutip_qobj(
    s_parameters_standard_matrix: jnp.ndarray,
):
    """
    This function converts the calculated S-parameters into a standard Unitary matrix topology so that the shape and
    dimensions of the matrix can be observed.

    I think this means we need to transpose the output of the filtered sax SDense matrix to map it to a QuTip matrix.
    Note that the documentation and formatting of the standard `sax` mapping to a S-parameter standard notation is
    already in described in piel/piel/sax/utils.py.

    From this stage we can implement a ``QObj`` matrix accordingly and perform simulations accordingly. https://qutip.org/docs/latest/guide/qip/qip-basics.html#unitaries

    For example, a ``qutip`` representation of an s-gate gate would be:

    ..code-block::

        import numpy as np
        import qutip
        # S-Gate
        s_gate_matrix = np.array([[1.,   0], [0., 1.j]])
        s_gate = qutip.Qobj(mat, dims=[[2], [2]])

    In mathematical notation, this S-gate would be written as:

    ..math::

        S = \\begin{bmatrix}
            1 & 0 \\\\
            0 & i \\\\
        \\end{bmatrix}

    Args:
        s_parameters_standard_matrix (nso.ndarray): A dictionary of S-parameters in the form of a SDict from `sax`.

    Returns:
        qobj_unitary (qutip.Qobj): A QuTip QObj representation of the S-parameters in a unitary matrix.

    """
    s_parameter_standard_matrix_numpy = np.asarray(s_parameters_standard_matrix)
    qobj_unitary = qutip.Qobj(s_parameter_standard_matrix_numpy)
    return qobj_unitary


def subunitary_selection_on_index(
    unitary_matrix: jnp.ndarray,
    rows_index: jnp.ndarray | tuple,
    columns_index: jnp.ndarray | tuple,
):
    """
    This function returns a unitary between the indexes selected, and verifies the indexes are valid by checking that
    the output matrix is also a unitary.

    TODO implement validation of a 2D matrix.
    """
    if type(rows_index) is tuple:
        rows_index = jnp.asarray(rows_index)

    if type(columns_index) is tuple:
        rows_index = jnp.asarray(columns_index)

    unitary_matrix_row_selection = unitary_matrix.at[rows_index, :].get()
    unitary_matrix_row_column_selection = unitary_matrix_row_selection.at[
        :, columns_index
    ].get()
    return unitary_matrix_row_column_selection


def subunitary_selection_on_range(
    unitary_matrix: jnp.ndarray,
    stop_index: tuple,
    start_index: Optional[tuple] = (0, 0),
):
    """
    This function returns a unitary between the indexes selected, and verifies the indexes are valid by checking that
    the output matrix is also a unitary.

    TODO implement validation of a 2D matrix.
    """
    start_row = start_index[0]
    end_row = stop_index[0]
    start_column = start_index[1]
    end_column = stop_index[1]
    column_range = jnp.arange(start_column, end_column + 1)
    row_range = jnp.arange(start_row, end_row + 1)
    unitary_matrix_row_column_selection = subunitary_selection_on_index(
        unitary_matrix=unitary_matrix,
        rows_index=row_range,
        columns_index=column_range,
    )
    return unitary_matrix_row_column_selection


def verify_matrix_is_unitary(matrix: jnp.ndarray) -> bool:
    """
    Verify that the matrix is unitary.

    Args:
        matrix (jnp.ndarray): The matrix to verify.

    Returns:
        bool: True if the matrix is unitary, False otherwise.
    """
    qobj = matrix_to_qutip_qobj(matrix)
    return qobj.check_isunitary()


standard_s_parameters_to_qutip_qobj = matrix_to_qutip_qobj



import hdl21 as h

__all__ = ["straight", "StraightParameters"]


@h.paramclass
class StraightParameters:
    """
    These are all the potential parametric configuration variables
    """

    pass


@h.generator
def straight(params: StraightParameters) -> h.Module:
    """
    Implements a `hdl21` taper resistor class.
    """

    @h.module
    class Straight:
        e1, e2 = h.Ports(2)
        r1 = h.IdealResistor(r=1e3)(p=e1, n=e2)

    return Straight



import hdl21 as h

__all__ = ["via_stack", "ViaStackParameters"]


@h.paramclass
class ViaStackParameters:
    pass


@h.generator
def via_stack(params: ViaStackParameters) -> h.Module:
    """
    Implements a `hdl21` taper resistor class. We need to include the mapping ports as we expect our gdsfactory component to be with the instance of the model.
    """

    @h.module
    class ViaStack:
        """
        This is unrealistic but need to PR to hdl21 for validitiy
        """

        e1, e2, e3, e4 = h.Ports(4)
        # TODO PR this valid to hdl21
        # All bottom ports connected together
        # e1 = e2 = e3 = e4
        # Top port e5 is output of via
        r1 = h.IdealResistor(r=1e3)(p=e1, n=e2)
        r2 = h.IdealResistor(r=1e3)(p=e2, n=e3)
        r3 = h.IdealResistor(r=1e3)(p=e3, n=e4)
        r4 = h.IdealResistor(r=1e3)(p=e4, n=e1)

    return ViaStack



import itertools

__all__ = [
    "single_parameter_sweep",
    "multi_parameter_sweep",
]


def single_parameter_sweep(
    base_design_configuration: dict,
    parameter_name: str,
    parameter_sweep_values: list,
):
    """
    This function takes a base_design_configuration dictionary and sweeps a single parameter over a list of values. It returns a list of dictionaries that correspond to the parameter sweep.

    Args:
        base_design_configuration(dict): Base design configuration dictionary.
        parameter_name(str): Name of parameter to sweep.
        parameter_sweep_values(list): List of values to sweep.

    Returns:
        parameter_sweep_design_dictionary_array(list): List of dictionaries that correspond to the parameter sweep.
    """
    parameter_sweep_design_dictionary_array = []
    for parameter in parameter_sweep_values:
        design = base_design_configuration.copy()
        design[parameter_name] = parameter
        parameter_sweep_design_dictionary_array.extend([design])
    return parameter_sweep_design_dictionary_array


def multi_parameter_sweep(
    base_design_configuration: dict, parameter_sweep_dictionary: dict
) -> list:
    """
    This multiparameter sweep is pretty cool, as it will generate designer list of dictionaries that comprise of all the possible combinations of your parameter sweeps. For example, if you are sweeping `parameter_1 = np.arange(0, 2) = array([0, 1])`, and `parameter_2 = np.arange(2, 4) = array([2, 3])`, then this function will generate list of dictionaries based on the default_design dictionary, but that will comprise of all the potential parameter combinations within this list.

    For the example above, there arould be 4 combinations [(0, 2), (0, 3), (1, 2), (1, 3)].

    If you were instead sweeping for `parameter_1 = np.arange(0, 5)` and `parameter_2 = np.arange(0, 5)`, the dictionary generated would correspond to these parameter combinations of::
        [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (1, 0), (1, 1), (1, 2), (1, 3), (1, 4), (2, 0), (2, 1), (2, 2), (2, 3), (2, 4), (3, 0), (3, 1), (3, 2), (3, 3), (3, 4), (4, 0), (4, 1), (4, 2), (4, 3), (4, 4)].

    Make sure to use the parameter_names from default_design when writing up the parameter_sweep dictionary key name.

    Example project_structure formats::

        example_parameter_sweep_dictionary = {
            "parameter_1": np.arange(1, -40, 1),
            "parameter_2": np.arange(1, -40, 1),
        }

        example_base_design_configuration = {
            "parameter_1": 10.0,
            "parameter_2": 40.0,
            "parameter_3": 0,
        }

    Args:
        base_design_configuration(dict): Dictionary of the default design configuration.
        parameter_sweep_dictionary(dict): Dictionary of the parameter sweep. The keys should be the same as the keys in the base_design_configuration dictionary.

    Returns:
        parameter_sweep_design_dictionary_array(list): List of dictionaries that comprise of all the possible combinations of your parameter sweeps.
    """
    parameter_names_sweep_list = []
    parameter_sweep_values_list = []
    parameter_sweep_design_dictionary_array = []

    for parameter_sweep_name in parameter_sweep_dictionary.keys():
        parameter_names_sweep_list.extend([parameter_sweep_name])
        parameter_sweep_values_list.extend(
            [parameter_sweep_dictionary[parameter_sweep_name].tolist()]
        )
    sweep_combinations = list(itertools.product(*parameter_sweep_values_list))

    for parameter_combination in sweep_combinations:
        design = base_design_configuration.copy()
        parameter_index = 0
        for parameter in parameter_combination:
            design[parameter_names_sweep_list[parameter_index]] = parameter
            parameter_index += 1
        parameter_sweep_design_dictionary_array.extend([design])

    return parameter_sweep_design_dictionary_array



import jax.numpy as jnp

__all__ = [
    "heat_transfer_1d_W",
]


def heat_transfer_1d_W(
    thermal_conductivity_fit, temperature_range_K, cross_sectional_area_m2, length_m
) -> float:
    thermal_conductivity_integral_area = jnp.trapz(
        thermal_conductivity_fit, temperature_range_K
    )
    return cross_sectional_area_m2 * thermal_conductivity_integral_area / length_m



from .defaults import get_default_models

__all__ = [
    "compose_custom_model_library_from_defaults",
]


def compose_custom_model_library_from_defaults(custom_models: dict) -> dict:
    """
    Compose the default models with the custom models.

    Args:
        custom_models (dict): Custom models dictionary.

    Returns:
        dict: Composed models dictionary.
    """
    return {**get_default_models(), **custom_models}


def check_is_unitary(model: dict) -> bool:
    """
    Check if the model is unitary.

    Args:
        model (dict): Model dictionary.

    Returns:
        bool: True if unitary, False if not.
    """
    # TODO
    pass



"""
This file allows us to automate several aspects of creating a fully compatible project structure.
"""
from .config import piel_path_types
from .file_system import return_path, write_script, read_json, check_path_exists

__all__ = [
    "read_configuration",
    "create_setup_py_from_config_json",
]


def read_configuration(design_directory: piel_path_types) -> dict:
    """
    This function reads the configuration file found in the design directory.

    Args:
        design_directory(piel_path_types): Design directory PATH.

    Returns:
        config_dictionary(dict): Configuration dictionary.
    """
    design_directory = return_path(design_directory)
    config_path = design_directory / "config.json"
    check_path_exists(config_path, raise_errors=True)
    config_dictionary = read_json(config_path)
    return config_dictionary


def create_setup_py_from_config_json(design_directory: piel_path_types) -> None:
    """
    This function creates a setup.py file from the config.json file found in the design directory.

    Args:
        design_directory(piel_path_types): Design directory PATH or module name.

    Returns:
        None
    """
    design_directory = return_path(design_directory)
    config_json_path = design_directory / "config.json"
    config_dictionary = read_json(config_json_path)
    commands_list = [
        "#!/usr/bin/env python",
        "from distutils.core import setup \n",
        "setup(name=" + config_dictionary["NAME"] + ",",
        "\tversion=" + str(config_dictionary["VERSION"]) + ",",
        "\tdescription=" + config_dictionary["DESCRIPTION"] + ","
        "\tauthor=" + "Dario Quintero" + ",",
        "\tauthor_email=" + "darioaquintero@gmail.com" + ",",
        "\turl=" + "https://github.com/daquintero/piel" + ",",
        "\tpackages=" + str(["models", "tb"]) + ",",
        ")",
    ]
    script = " \n".join(commands_list)
    write_script(directory_path=design_directory, script=script, script_name="setup.py")



"""
Translated from https://github.com/flaport/sax or https://github.com/flaport/photontorch/tree/master
"""
import sax

__all__ = ["mmi1x2_50_50"]


def mmi1x2_50_50():
    S = {
        ("o1", "o2"): 0.5**0.5,
        ("o1", "o3"): 0.5**0.5,
    }
    return sax.reciprocal(S)



import jax.numpy as jnp
import time
import thewalrus
import numpy as np

__all__ = ["unitary_permanent"]


def unitary_permanent(
    unitary_matrix: jnp.ndarray,
) -> tuple:
    """
    The permanent of a unitary is used to determine the state probability of combinatorial Gaussian boson samping systems.

    ``thewalrus`` Ryser's algorithm permananet implementation is described here: https://the-walrus.readthedocs.io/en/latest/gallery/permanent_tutorial.html

    Note that this function needs to be as optimised as possible, so we need to minimise our computational complexity of our operation.

    # TODO implement validation
    # TODO maybe implement subroutine if computation is taking forever.
    # TODO why two outputs? Understand this properly later.

    Args:
        unitary_permanent (np.ndarray): The unitary matrix.

    Returns:
        tuple: The circuit permanent and the time it took to compute it.

    """
    start_time = time.time()
    unitary_matrix_numpy = np.asarray(unitary_matrix)
    circuit_permanent = thewalrus.perm(unitary_matrix_numpy)
    end_time = time.time()
    computed_time = end_time - start_time
    return circuit_permanent, computed_time



# TODO finish
from .photonic import straight_waveguide
from .photonic import mmi1x2
from .photonic import mmi2x2
from .photonic import coupler_simple

__all__ = [
    "get_all_models",
]

# Default model dictionary library that can be overwritten for specific modelling applications.
__default_models_dictionary__ = {
    "coupler": coupler_simple,
    "mmi1x2_50_50": mmi1x2.mmi1x2_50_50,
    "mmi2x2_50_50": mmi2x2.mmi2x2_50_50,
    "ideal_active_waveguide": straight_waveguide.ideal_active_waveguide,
    "simple_straight": straight_waveguide.simple_straight,
    "waveguide": straight_waveguide.waveguide,
}


def get_all_models(custom_library: dict | None = None) -> dict:
    """
    Returns the default models dictionary.

    Args:
        custom_library (dict): Custom defaults dictionary.

    Returns:
        dict: Default models dictionary.
    """
    if custom_library is not None:
        return custom_library
    else:
        return __default_models_dictionary__



from .geometry import *
from .units import *

from piel.models.physical import photonic  # NOQA: F401
from piel.models.physical import electrical  # NOQA: F401
from piel.models.physical import electronic  # NOQA: F401
from piel.models.physical import electro_optic  # NOQA: F401
from piel.models.physical import opto_electronic  # NOQA: F401



from piel.models.transient import photonic  # NOQA: F401
from piel.models.transient import electrical  # NOQA: F401
from piel.models.transient import electronic  # NOQA: F401
from piel.models.transient import electro_optic  # NOQA: F401
from piel.models.transient import opto_electronic  # NOQA: F401



from piel.models.logic import photonic  # NOQA: F401
from piel.models.logic import electrical  # NOQA: F401
from piel.models.logic import electronic  # NOQA: F401
from piel.models.logic import electro_optic  # NOQA: F401
from piel.models.logic import opto_electronic  # NOQA: F401



from itertools import product
import math
import numpy as np
from typing import Optional
import qutip

__all__ = [
    "all_fock_states_from_photon_number",
    "fock_state_nonzero_indexes",
    "fock_state_to_photon_number_factorial",
    "fock_states_at_mode_index",
]


def all_fock_states_from_photon_number(
    mode_amount: int,
    photon_amount: int = 1,
) -> list[qutip.Qobj]:
    """
    For a specific amount of modes, we can generate all the possible Fock states for whatever amount of input photons we desire. This returns a list of all corresponding Fock states.

    Args:
        mode_amount (int): The amount of modes in the system.
        photon_amount (int, optional): The amount of photons in the system. Defaults to 1.

    Returns:
        list: A list of all the Fock states.
    """
    photon_numbers = [[i] for i in range(photon_amount + 1)]
    states = []

    for photon_number in product(photon_numbers, repeat=mode_amount):
        state_values = np.array(photon_number).reshape(mode_amount, 1)
        state = qutip.Qobj(state_values)
        states.append(state)
    return states


def fock_state_to_photon_number_factorial(fock_state: qutip.Qobj) -> float:
    """
    This function converts a Fock state defined as:

    .. math::
        \newcommand{\ket}[1]{\left|{#1}\right\rangle}
        \ket{f_1} = \ket{j_1, j_2, ... j_N}$

    and returns:

    .. math::

        j_1^{'}! j_2^{'}! ... j_N^{'}!

    Args:
        fock_state (qutip.Qobj): A QuTip QObj representation of the Fock state.

    Returns:
        float: The photon number factorial of the Fock state.
    """
    # TODO implement checks of Fock state validity
    photon_number_factorial = 1
    for photon_number in fock_state:
        photon_number_amount = photon_number[0].real
        photon_number_factorial *= math.factorial(int(photon_number_amount[0]))
    return photon_number_factorial


def fock_state_nonzero_indexes(fock_state: qutip.Qobj) -> tuple[int]:
    """
    This function returns the indexes of the nonzero elements of a Fock state.

    Args:
        fock_state (qutip.Qobj): A QuTip QObj representation of the Fock state.

    Returns:
        tuple: The indexes of the nonzero elements of the Fock state.
    """
    # TODO implement checks of Fock state validity
    nonzero_indexes = []
    for index, photon_number in enumerate(fock_state):
        photon_number_amount = photon_number[0].real
        if photon_number_amount[0] != 0:
            nonzero_indexes.append(index)
    return tuple(nonzero_indexes)


def fock_states_at_mode_index(
    mode_amount: int,
    target_mode_index: int,
    maximum_photon_amount: Optional[int] = 1,
) -> list[qutip.Qobj]:
    """
    This function returns a list of valid Fock states that fulfill a condition of having a maximum photon number at a specific mode index.

    Args:
        mode_amount (int): The amount of modes in the system.
        target_mode_index (int): The mode index to check the photon number at.
        maximum_photon_amount (int, optional): The amount of photons in the system. Defaults to 1.

    Returns:
        list: A list of all the Fock states.
    """

    def check_photon_number_at_mode(state_value: np.ndarray) -> bool:
        # Check if mode_index is valid
        if target_mode_index < len(state_value):
            # Return comparison result
            return (state_value[target_mode_index][0] <= maximum_photon_amount) and (
                state_value[target_mode_index][0] > 0
            )
        else:
            # Index out of range.
            return False

    photon_numbers = [[i] for i in range(maximum_photon_amount + 1)]
    states = []

    for photon_number in product(photon_numbers, repeat=mode_amount):
        state_values = np.array(photon_number).reshape(mode_amount, 1)
        if check_photon_number_at_mode(state_values):
            state = qutip.Qobj(state_values)
            states.append(state)
    return states



"""
These set of functions aim to provide functionality to automate interacting with OpenLanes v1 design into Python environment, whilst `OpenLanes2` is under development.
"""

import os
import pathlib
import json
from piel.parametric import multi_parameter_sweep
from piel.file_system import (
    copy_source_folder,
    permit_script_execution,
    read_json,
    return_path,
    run_script,
    write_script,
)

__all__ = [
    "check_config_json_exists_openlane_v1",
    "check_design_exists_openlane_v1",
    "configure_and_run_design_openlane_v1",
    "configure_parametric_designs_openlane_v1",
    "configure_flow_script_openlane_v1",
    "create_parametric_designs_openlane_v1",
    "get_design_directory_from_root_openlane_v1",
    "get_latest_version_root_openlane_v1",
    "read_configuration_openlane_v1",
    "write_configuration_openlane_v1",
]


def check_config_json_exists_openlane_v1(
    design_name: str,
    root_directory: str | pathlib.Path | None = None,
) -> bool:
    """
    Checks if a design has a `config.json` file.

    Args:
        design_name(str): Name of the design.

    Returns:
        config_json_exists(bool): True if `config.json` exists.
    """
    config_json_exists = False
    design_directory = get_design_directory_from_root_openlane_v1(
        design_name=design_name, root_directory=root_directory
    )
    if (design_directory / "config.json").exists():
        config_json_exists = True
    return config_json_exists


def check_design_exists_openlane_v1(
    design_name: str,
    root_directory: str | pathlib.Path | None = None,
) -> bool:
    """
    Checks if a design exists in the OpenLane v1 design folder.

    Lists all designs inside the Openlane V1 design root.

    Args:
        design_name(str): Name of the design.

    Returns:
        design_exists(bool): True if design exists.
    """
    if root_directory is None:
        root_directory = get_latest_version_root_openlane_v1()

    design_exists = False
    openlane_v1_design_directory = root_directory / "designs"
    all_openlane_v1_designs = list(openlane_v1_design_directory.iterdir())
    if (openlane_v1_design_directory / design_name) in all_openlane_v1_designs:
        design_exists = True
    return design_exists


def configure_and_run_design_openlane_v1(
    design_name: str,
    configuration: dict | None = None,
    root_directory: str | pathlib.Path | None = None,
) -> None:
    """
    Configures and runs an OpenLane v1 design.

    This function does the following:
    1. Check that the design_directory provided is under $OPENLANE_ROOT/<latestversion>/designs
    2. Check if `config.json` has already been provided for this design. If a configuration dictionary is inputted into the function parameters, then it overwrites the default `config.json`.
    3. Create a script directory, a script is written and permissions are provided for it to be executable.
    4. Permit and execute the `openlane_flow.sh` script in the `scripts` directory.

    Args:
        design_name(str): Name of the design.
        configuration(dict | None): Configuration dictionary.
        root_directory(str | pathlib.Path): Design directory.

    Returns:
        None
    """

    design_directory = get_design_directory_from_root_openlane_v1(
        design_name=design_name, root_directory=root_directory
    )

    # Check configuration
    config_json_exists = check_config_json_exists_openlane_v1(design_name)

    if config_json_exists:
        pass
    else:
        if configuration is None:
            raise ValueError(
                "Configuration dictionary is None. Please provide a configuration dictionary."
            )
        else:
            write_configuration_openlane_v1(
                configuration=configuration, design_directory=design_directory
            )

    # Create script directory
    configure_flow_script_openlane_v1(design_name=design_name)

    # Execute script
    openlane_flow_script_path = design_directory / "scripts" / "openlane_flow.sh"
    permit_script_execution(openlane_flow_script_path)
    run_script(openlane_flow_script_path)


def configure_flow_script_openlane_v1(
    design_name: str,
    root_directory: str | pathlib.Path | None = None,
) -> None:
    """
    Configures the OpenLane v1 flow script after checking that the design directory exists.

    Args:
        design_directory(str | pathlib.Path | None): Design directory. Defaults to latest OpenLane root.

    Returns:
        None
    """
    design_directory = get_design_directory_from_root_openlane_v1(
        design_name=design_name, root_directory=root_directory
    )
    if check_design_exists_openlane_v1(design_name):
        commands_list = [
            "#!/bin/sh",
            "cd " + str(root_directory),
            "./flow.tcl -design " + design_name,
        ]
        script = " \n".join(commands_list)
        write_script(
            directory_path=design_directory / "scripts",
            script=script,
            script_name="openlane_flow.sh",
        )
    else:
        raise ValueError(
            "Design: "
            + str(design_name)
            + " not found in "
            + os.environ["OPENLANE_ROOT"]
        )


def configure_parametric_designs_openlane_v1(
    design_name: str,
    parameter_sweep_dictionary: dict,
    add_id: bool = True,
) -> list:
    """
    For a given `source_design_directory`, this function reads in the config.json file and returns a set of parametric sweeps that gets used when creating a set of parametric designs.

    Args:
        add_id(bool): Add an ID to the design name. Defaults to True.
        parameter_sweep_dictionary(dict): Dictionary of parameters to sweep.
        source_design_directory(str | pathlib.Path): Source design directory.

    Returns:
        configuration_sweep(list): List of configurations to sweep.
    """
    source_configuration = read_configuration_openlane_v1(design_name=design_name)
    configuration_sweep = multi_parameter_sweep(
        base_design_configuration=source_configuration,
        parameter_sweep_dictionary=parameter_sweep_dictionary,
    )
    if add_id:
        i = 0
        for configuration_i in configuration_sweep:
            # Checks the unique ID of the configuration
            configuration_id = id(configuration_i)
            # Adds the ID to the configuration list
            configuration_sweep[i]["id"] = configuration_id
            i += 1
    return configuration_sweep


def create_parametric_designs_openlane_v1(
    design_name: str,
    parameter_sweep_dictionary: dict,
    target_directory: str | pathlib.Path | None = None,
) -> None:
    """
    Takes a OpenLane v1 source directory and creates a parametric combination of these designs.

    Args:
        design_name(str): Name of the design.
        parameter_sweep_dictionary(dict): Dictionary of parameters to sweep.
        target_directory(str | pathlib.Path | None): Optional target directory.

    Returns:
        None
    """
    source_design_directory = get_design_directory_from_root_openlane_v1(
        design_name=design_name
    )
    source_design_name = design_name

    if target_directory is None:
        target_directory = get_latest_version_root_openlane_v1() / "designs"

    parameter_sweep_configuration_list = configure_parametric_designs_openlane_v1(
        add_id=True,
        design_name=design_name,
        parameter_sweep_dictionary=parameter_sweep_dictionary,
    )

    for configuration_i in parameter_sweep_configuration_list:
        # Create a target directory with the name of the design and the configuration ID
        target_directory_i = (
            target_directory / source_design_name + "_" + str(configuration_i["id"])
        )
        # Copy the source design directory to the target directory
        copy_source_folder(
            source_directory=source_design_directory,
            target_directory=target_directory_i,
        )


def get_latest_version_root_openlane_v1() -> pathlib.Path:
    """
    Gets the latest version root of OpenLane v1.
    """
    openlane_tool_directory = pathlib.Path(os.environ["OPENLANE_ROOT"])
    latest_openlane_version = list(openlane_tool_directory.iterdir())
    openlane_v1_design_directory = openlane_tool_directory / latest_openlane_version[-1]
    return openlane_v1_design_directory


def get_design_directory_from_root_openlane_v1(
    design_name: str,
    root_directory: str | pathlib.Path | None = None,
) -> pathlib.Path:
    """
    Gets the design directory from the root directory.

    Args:
        design_name(str): Name of the design.
        root_directory(str | pathlib.Path): Design directory.

    Returns:
        design_directory(pathlib.Path): Design directory.
    """
    if root_directory is None:
        root_directory = get_latest_version_root_openlane_v1()

    root_directory = return_path(root_directory)
    design_exists = check_design_exists_openlane_v1(design_name)
    if design_exists:
        pass
    else:
        raise ValueError(
            "Design: "
            + design_name
            + " is not found in "
            + str(root_directory / "designs")
        )
    design_directory = root_directory / "designs" / design_name
    return design_directory


def read_configuration_openlane_v1(
    design_name: str,
    root_directory: str | pathlib.Path | None = None,
) -> dict:
    """
    Reads a `config.json` from a design directory.

    Args:
        design_name(str): Design name.
        root_directory(str | pathlib.Path): Design directory.

    Returns:
        configuration(dict): Configuration dictionary.
    """
    config_json_exists = check_config_json_exists_openlane_v1(design_name)
    if config_json_exists:
        design_directory = get_design_directory_from_root_openlane_v1(
            design_name=design_name, root_directory=root_directory
        )
        configuration = read_json(design_directory / "config.json")
        return configuration
    else:
        raise ValueError(
            "Configuration file for design: "
            + design_name
            + " is not found in "
            + str(root_directory / "designs" / design_name)
        )


def write_configuration_openlane_v1(
    configuration: dict,
    design_directory: str | pathlib.Path,
) -> None:
    """
    Writes a `config.json` onto a `design_directory`

    Args:
        configuration(dict): OpenLane configuration dictionary.
        design_directory(str): Design directory PATH.

    Returns:
        None
    """
    with open(str((design_directory / "config.json").resolve()), "w") as write_file:
        json.dump(configuration, write_file, indent=4)
